//                              -*- Mode: C++ -*-
//: EMClusterMember.C
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 3/4/97
// Last Update ........
//

#include "care/EMClusterMember.hh"
#include "care/EMCluster.hh"
#include "care/EAClusterAnalyzer.hh"
#include"care/EDClusterModifier.hh"
#include "arte/ArteMessage.hh"

#undef WRITE
/*
CLASS
  EMClusterMember
  Abstract base class for the hierarchical cluster tree used in
  pattern recognition of ECAL

class EMClusterMember {
public:
  //-- Constructors
  //   The construction of EMClusterMember is done referring to some
  //   parent. The depth level is adjusted accordingly.
  EMClusterMember(const EMClusterMember*  Parent);

  // Status access functions
  //--
  // Return a flag to tell if this EMClusterMember is a leaf (if it is not, then
  // it can be a clustermember list
  virtual bool IsALeaf() const;

  //--
  // Return the maximum energy measured in the cluster's (and sub clusters) 
  // cells or the Energy of the Cell itself. 
  virtual Energy_t  MaxEnergy() const;
  //--
  // After operation on cluster such as reordering, merging or other
  // It is possible that the depths are not correct. With this function
  // the depth of the EMClusterMember and all its children are adjusted. 
  virtual void Adjust_depths(int depth);
  //--
  // Analysis and traversing of the cluster tree is done by EAClusterAnalyzers
  // that are object with definite analyzing properties. 
  // EMClusterMembers MUST ignore these properties: in this way a loose coupling
  // between EMClusterMembers and EAClusterAnalyzers is achieved and both
  // can vary independently during program evolution.
  // Examples of EAClusterAnalyzers are those that should return
  // total energy of a cluster, the Center of Gravity, the Number of Cells.. 
  virtual void AnalyzeMe(EAClusterAnalyzer & Analyzer);
protected:
  //--
  //  The default constructors are protected in order to prevent
  //  wrong usage of this class by user.
  EMClusterMember();
  // Data part
  //-- 
  //  This is a link to the parent of the current node.
  //  If there is no parent its value is NULL.
  EMClusterMember*  ParentCluster;
  //--
  //  This is a deepness level for fast evaluation in analysis.
  //  Is it usefull?
  int  depth;
private:
};
*/



//////////////  Class EMClusterMember definitions ///////////////////////////
//-- Constructors
//   The construction of EMClusterMember is done referring to some
//   parent. The depth level is adjusted accordingly.

EMClusterMember::EMClusterMember( EMCluster* parent) :
  parentCluster(parent), nextClusterMember(0), previousClusterMember(0) {
    if(parent->lastChild!=0){
      previousClusterMember = parent->lastChild;
      previousClusterMember->nextClusterMember = this;
    } else {
      parent->lastChild = this;
      parent->firstChild = this;
    }
    depth = (Level_t) (parent->depthLevel() +1);
}

EMClusterMember::EMClusterMember() :
  parentCluster(0), nextClusterMember(0), 
  previousClusterMember(0),depth(TOPLEVEL) {}

EMClusterMember::~EMClusterMember() {}
//--
// After operation on cluster such as reordering, merging or other
// It is possible that the depths are not correct. With this function
// the depth of the EMClusterMember and all its children are adjusted. 
void EMClusterMember::adjust_depths() {
  depth = (Level_t) (parentCluster->depthLevel() +1);
}

//--
// Add a Child to the list: it makes sense only for cluster objects
EMClusterMember*  EMClusterMember::add(EMClusterMember * child){
  arteMsg<<" Error: attempting to add a Child to a non Cluster object! ";
  arteMsg.warn();
  arteMsg<<" Adding the element to the parent of this Child instead.";
  arteMsg.warn();
  return parentCluster->add(child);
} 


//--
// Analysis and traversing of the cluster tree is done by EAClusterAnalyzers
// that are object with definite analyzing properties. 
// EMClusterMembers MUST ignore these properties: in this way a loose coupling
// between EMClusterMembers and EAClusterAnalyzers is achieved and both
// can vary independently during program evolution.
// Examples of EAClusterAnalyzers are those that should return
// total energy of a cluster, the Center of Gravity, the Number of Cells.. 
void EMClusterMember::analyzeMe(EAClusterAnalyzer & analyzer) const {
  analyzer.doAnalysis(this);
}

void EMClusterMember::modifyMe(EDClusterModifier & modifier) {
  modifier.doModify(this);
}




//--
//  Dump for debugging
void EMClusterMember::dump() const{
  arteMsg<<" Error: Dumping a EMClusterMember object !";
  arteMsg.warn();
}

//--
// Append the current object to the list defined 
// by the parent cluster
EMClusterMember* EMClusterMember::appendTo(EMCluster* parent){
# ifdef WRITE
  arteMsg<<" Appending current EMClusterMember to a parent of depth "
	 << parent->depthLevel();
  arteMsg.allp();
# endif
  nextClusterMember = 0; // this object is the last in the list
  previousClusterMember = parent->lastChild;   // get the previous in the list
  if(parent->lastChild!=0){
    parent->lastChild->nextClusterMember = this; // update forward link 
                                               // of the previous child
  } else {
    parent->firstChild = this;
  }
  parentCluster = parent;                      // set the parent cluster
  parent->lastChild = this;            // Update parent link to last child
  adjust_depths(); // update depths
  return this;
} 
 

//--
//  Follow upstream all the links until level "lev" is reached
//  return 0 or the Cluster* in which level "lev" is
EMCluster* EMClusterMember::upstream(Level_t level) const{
  //  
  EMCluster* parent = parentCluster;
  while( parent!=0 && parent->depthLevel()>level){
    parent = parent->getParent();
  }
  if(level>parent->depthLevel())
    parent = 0;
  return parent;
}









