//                              -*- Mode: C++ -*-
//:EcalDB.C
//
//
// Author: Vincenzo Vagnoni & Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
//

#include "care/EcalDB.hh"

// ------------ Static data definitions

char *EcalDB::tableName[3]  = { "ECBO", "ECHS", "ECHF" };
char *EcalDB::tableDesc[3] = {
  // ECBO description string for Mizzi
  "I:INFO-COLUMN I:FDID I:STAT I:NUMB I:LINK I:CHAN",
  // ECHS description string for Mizzi
  "I:INFO-COLUMN F:PEDS F:PCAL F:ECAL F:THR1 F:THR2 F:MIP F:VMIP F:CHST"
  " I:FDID I:CHAN I:ECBO",
  // ECHF description string for Mizzi
  "I:INFO-COLUMN I:STAT F:PEDF F:WID F:Q"
};
char *EcalDB::tableComment[3]= {
    "Ecal readout BOard data",
    "Ecal readout CHannel Calibration table (Slow) - Slow update rate",
    "Ecal readout CHannel Calibration table (Fast) - Fast update rate"
  };
int   EcalDB::fieldDim[3][11] = {
    // ECBO field dimensions
    { 1, 1, 1, 1, 32 },
    // ECHS field dimensions
    { 1, 6, 6, 1, 1, 1, 1, 1, 1, 1, 1 },
    // ECHF field dimensions
    { 1, 1, 1, 1 }
  };

int   EcalDB::fieldNum[3] = { 5, 11, 4 };
int   EcalDB::fieldType[3][11]  = {
    // ECBO field types
    { TABLEint, TABLEint, TABLEint, TABLEint, TABLEint },
    // ECHS field types
    { TABLEfloat, TABLEfloat, TABLEfloat, TABLEfloat,
      TABLEfloat, TABLEfloat, TABLEfloat, TABLEfloat,
      TABLEint,   TABLEint,   TABLEint },
    // ECHF field types
    { TABLEint, TABLEfloat, TABLEfloat, TABLEfloat }
  };


EcalDB::EcalDB() : ECBO(0,*this), ECHS(1,*this), ECHF(2,*this) {};

EcalDB::~EcalDB() {};


Tables EcalDB::Open(char *dbName, int access) {
  x=OpenTables(dbName, access);
  return x;
};

int EcalDB::Close(void) {
  int ret=CloseTables(x);
  x=NULL;
  return ret;
};





EcalDB::TableMeth::TableMeth
(int i, EcalDB &T) : xeCur(i), EcalDBase(T), tSize(0) {

  for(int j=0; j<EcalDBase.fieldNum[xeCur]; j++) {
    tSize += EcalDBase.fieldDim[xeCur][j]*4;
  }

};

EcalDB::TableMeth::~TableMeth() {};

int EcalDB::TableMeth::Book() {

  EcalDBase.xe[xeCur]=
    BookTable(EcalDBase.x,
	      EcalDBase.tableName[xeCur],
	      EcalDBase.tableDesc[xeCur],
	      EcalDBase.tableComment[xeCur]);
  return EcalDBase.xe[xeCur];

}

int EcalDB::TableMeth::Unbook() {
  int ret;
  ret=UnbookTable(EcalDBase.x,EcalDBase.xe[xeCur]);
  EcalDBase.xe[xeCur]=0;
  return ret;
}

int EcalDB::TableMeth::Store
(int nrow, void *ptr, int maj, int min, int access) {

  char *table=(char*)ptr;

  TableStamp(EcalDBase.x,maj,min);
  SkipTableField(EcalDBase.x,1);
  PutTableField(EcalDBase.x,TABLEint,1,&nrow);

  for(int i=0; i<fieldNum[xeCur]; i++) {
    int dim=fieldDim[xeCur][i]*4;
    char *buffer=new char[dim*nrow];
    int bufferCursor=0;
    int tableCursor=0;
    for(int j=0; j<nrow; j++) {
      memcpy(&buffer[bufferCursor],&table[tableCursor],dim);
      bufferCursor+=dim;
      tableCursor+=tSize;
    }

    PutTableField(EcalDBase.x,EcalDBase.fieldType[xeCur][i],nrow*dim/4,buffer);
    delete buffer;

    table+=dim;
  }

  if(!StoreTable(EcalDBase.x,EcalDBase.xe[xeCur],access)) return 0;
  major=TableMajor(EcalDBase.x);
  minor=TableMinor(EcalDBase.x);

  return 1;
}

int EcalDB::TableMeth::Nrow(int maj, int min, int access) {
  TableStamp(EcalDBase.x,maj,min);
  if(!FetchTable(EcalDBase.x,EcalDBase.xe[xeCur],access)) return 0;

  SkipTableField(EcalDBase.x,1);
  GetTableField(EcalDBase.x,TABLEint,1,&nrow);
  return nrow;
}

int EcalDB::TableMeth::Fetch(void *ptr) {

  char *table=(char*)ptr;

  for(int i=0; i<fieldNum[xeCur]; i++) {
    int dim=fieldDim[xeCur][i]*4;
    char *buffer=new char[dim*nrow];
    int bufferCursor=0;
    int tableCursor=0;

    GetTableField(EcalDBase.x,EcalDBase.fieldType[xeCur][i],nrow*dim/4,buffer);

    for(int j=0; j<nrow; j++) {
      memcpy(&table[tableCursor],&buffer[bufferCursor],dim);
      bufferCursor+=dim;
      tableCursor+=tSize;
    }


    delete buffer;

    table+=dim;
  }
  major=TableMajor(EcalDBase.x);
  minor=TableMinor(EcalDBase.x);
  return 1;

}

int EcalDB::TableMeth::Match(int maj, int min, int access) {
  TableStamp(EcalDBase.x,maj,min);
  if(!MatchTable(EcalDBase.x,EcalDBase.xe[xeCur],access)) return 0;
  major=TableMajor(EcalDBase.x);
  minor=TableMinor(EcalDBase.x);
  return 1;
}

int EcalDB::TableMeth::Major() {
  return major;
}

int EcalDB::TableMeth::Minor() {
  return minor;
}
