
// *****************************************************************************
//
// source:    faRead.C
//
// created:   15 Mar 96
// updated:   29 Mar 99 Andreas Gellrich
//
// Author:    Andreas Gellrich
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: Andreas.Gellrich@desy.de
//
// changes:   delayed reading (AG: 26-Nov-98)
//            fill exp number form DAQ record (AG: 26-Nov-98)
//            max number of slbs changed (AG: 26-Nov-98)
//            multiple file input (AG: 27-Nov-98)
//            bx number from DAQ record (AG: 27-Nov-98)
//            bug when filling event number in  emg mode cured (AG: 11 Dec 98)
//            removed hbook writing (AG: 14 Dec 99)
//            obey ANSI rules (AG: 15 Jan 99)
//            use sltevt exclusively for event numbering (AG: 15 Jan 99)
//            some more polishing (AG: 21 Jan 99)
//            avoid memory leak (AG: 27 Jan 99)
//            farm_node filling (AG: 10 Feb 99)
//            filling daq time into EVHD (AG: 10 Feb 99)
//            introduce handler for online event display (AG: 12 Feb 99)
//            removed Arte include (AG: 23 Feb 99)
//            removed faUtil.hh (AG: 23 Feb 99)
//            polishing (AG: 2 Mar 99)
//            polishing (AG: 11 Mar 99)
//            polishing (AG: 29 Mar 99)
//
// *****************************************************************************

// includes
#include "farm/fa.hh"

// C prototpyes
extern "C" void mmbook_( int & , const int & );
extern "C" void mmfree_( int & );

// externals
extern int  FAON;    // farm online flag
extern int  FADEB;   // debug flag
extern int  NOSPAR;  // sparsification flag
extern int  SEG_LEN; // event buffer size
extern char *evtBuf; // event buffer
extern int  nodeID;  // 4LT node ID

// *****************************************************************************
//
// name:      faread_
//
// purpose:   read event data to Arte tables
//
// called by: mmrevt_()
//
// input:     void
// return:    Arte error code
//
// hints:     HERA-B DAQ native format is BIG_ENDIAN (RISC)
//
// tables:    EVHD clear format
//            DIHD clear format
//
//            Dnnn byte stream (DAQ,SVD,OTR, ...)           
//
// *****************************************************************************

//
// program code
//
int
faread_()
{
  // inits
  int        rc    = GRR_OK;
  static int count = 0;
  count++;

  // cycling
  faCycle();

  //
  // read event to buffer
  //

  // read
  if( FAON ){
    // read emg
    if( ( rc = faReadEmg() ) != GRR_OK ){
      return ARTE_END;
    }
  }
  else {
    // read file
    if( ( rc = faReadFile() ) != ETB_ok ){
      return ARTE_END; // try next file
    }
  }
  
  // event header from buffer
  EVENT_HDR evtHdr;
  ESP_get_event_hdr( (EVENT_HDR *) evtBuf, &evtHdr );
#ifdef _DEB_
  if( FADEB ){
    cout << endl;
    cout << "********** READ EVENT: *****************************" << endl;
    cout << "length="            << evtHdr.length          << " "
	 << "recid=0x"    << hex << evtHdr.recid    << dec << " "
	 << "format=0x"   << hex << evtHdr.format   << dec << " "
	 << "sltevt="            << evtHdr.sltevt          << " "
	 << "compmask=0x" << hex << evtHdr.compmask << dec << " "
	 << "evtmask=0x"  << hex << evtHdr.evtmask  << dec << endl;
    cout << "****************************************************" << endl;
    cout << endl;
  }
#endif

  //
  // EVHD table
  //
  mmbook_( ArtePointer<EVHD>::getLink().getRef(), 1 );
  ArtePointer<EVHD> evhd( 1 );
  evhd->evt     = arteflag_.numevt = evtHdr.sltevt; // event number filled by l4s
  evhd->run     = arteflag_.numrun = 4711; // default run number
  evhd->exp     = arteflag_.numexp = 9999; // default exp number
  evhd->date    = faDate( (time_t *) 0 ); // current date
  evhd->time    = faTime( (time_t *) 0 ); // current time
  evhd->vfld    = 0; // magnet field will be filled later ...
  evhd->flagdst = 0; // event classification ...
  evhd->flag    = 0; // event classification ...

  // 
  // DIHD table
  //
  mmbook_( ArtePointer<DIHD>::getLink().getRef(), 1 );
  ArtePointer<DIHD> dihd( 1 );
  dihd->bx   = 0x0; // DAQ record
  dihd->flt  = 0x0; // DAQ record
  dihd->slt  = evtHdr.sltevt; // swapping done
  dihd->slb  = 0x0; // SLB_HDR
  dihd->cmsk = evtHdr.compmask; // swapping done
  dihd->emsk = evtHdr.evtmask;  // swapping done

  // DSVD (needs collection of FED blocks first)
  int  nDSVD   = 50000;
  int  lenDSVD = 0;
  char *pDSVD  = (char *) malloc( nDSVD );

  // inits
  const int maxSlb = 1000; // maximal SLB records
  SLB_HDR   *slbp[maxSlb];       // array of SLB records

  // get pointers to SLBs of COMP_ID
  int m = 0;
  if( ( rc = ETB_get_slb_points( (EVENT_HDR *) evtBuf, -1, maxSlb, slbp, &m ) ) == ETB_ok ){
#ifdef _DEB_
    if( FADEB ){
      cout << "ETB_get_slb_points() m=" << m << endl;
    }
#endif
  }
  else {
#ifdef _ERR_
    cerr << "ERROR: "
	 << "ETB_get_slb_points() rc=" << rc << endl;
#endif
    return ARTE_ERR;
  }
  
  // SLB pointer loop
  for( int i=0; i<m; i++ ){
    SLB_HDR slbHdr;
    ESP_get_slb_hdr( (SLB_HDR *) slbp[i], &slbHdr ); // translate SLB_HDR
#ifdef _DEB_
    if( FADEB ){
      cout << endl;
      cout << "READ SLB record: "                        << endl;
      cout << "i="               << i                    << " "
	   << "length="          << slbHdr.length        << " "
	   << "recid=0x"  << hex << slbHdr.recid  << dec << " "
	   << "fltevt=0x" << hex << slbHdr.fltevt << dec << " "
	   << "compid=0x" << hex << slbHdr.compid << dec << " "
	   << "slbid=0x"  << hex << slbHdr.slbid  << dec << " "
	   << "nodeid=0x" << hex << slbHdr.nodeid << dec << endl;
    }
#endif
    
    // table
    int  recid  = slbHdr.recid;
    int  compid = slbHdr.compid;
    int  slbid  = slbHdr.slbid;
    int  iName  = faID2Name( compid );
    int  size   = slbHdr.length - sizeof( SLB_HDR ) - sizeof( SLB_TAIL );
    char *pData = (char *) slbp[i] + sizeof( SLB_HDR  );

#ifdef _DEB_
    if( FADEB ){
      cout << "  "
	   << "compid=0x" << hex << compid   << dec << " "
	   << "slbid=0x"  << hex << slbid    << dec << " "
	   << "iName="    << iName                  << " "
	   << "name="     << faName( iName )        << " "
	   << "recid=0x"  << recid           << dec << " "
	   << "pData()="  << (void *) pData         << " " 
	   << "size="     << size                   << endl;
    }
#endif

    if( count < 11 ){
      if( size == 0 ){
#ifdef _OUT_
	cout << "  "
	     << count << " "
	     << "WARNING (message will stop after 10 events!!!): SLB record has NO data: "
	     << "compid=0x" << hex << compid << dec << " "
	     << "slbid=0x"  << hex << slbid  << dec << endl;
#endif
      }
    }

    //
    // handle SLB recods
    //
    int nSparse = 0;

    //
    // DDAQ 0x0000 *************************************************************
    //
    if( iName == faName2Int( "DDAQ" ) ){

      // pointer structure
      EVT_DATA *daq = (EVT_DATA *) pData;

      // fill 4LT node id to DDAQ record if not done before
      if( !daq->farm_node ){
	ESP_put_int( &nodeID, &( daq->farm_node ), 1 );
      }

      // get datime and put it to EVHD
      time_t datime = 0;
      ESP_get_int( &( daq->event_time ), (int *) &datime, 1 );
      evhd->date = faDate( (time_t *) &datime ); // current date
      evhd->time = faTime( (time_t *) &datime ); // current time
      
#ifdef _DEB_
      if( FADEB ){
	cout << "  DDAQ: "
	     << "nrow=" << ArtePointer<DDAQ>::nrow() << " "
	     << "size=" << size                      << endl;
      }
#endif
      mmbook_( ArtePointer<DDAQ>::getLink().getRef(), size / sizeof( int ) + 1 );
      ArtePointer<DDAQ> t( 1 ); // table
      int  *newID = (int *)   &t->bits; // new recID    
      char *p     = (char *)( &t->bits + 1 ); // data
      nSparse     = faCopySlb( &recid, pData, newID, p, size ); // copy
#ifdef _DEB_
      if( FADEB ){
	cout << "  DDAQ: "
	     << "recID=0x" << hex << *newID   << dec << " "
	     << "nrow()="         << t.nrow()        << " "
	     << "nSparse="        << nSparse         << endl;
      }
#endif
    }

    //
    // DSVD 0x0100 (first collect all FED blocks) ******************************
    //
    if( iName == faName2Int( "DSVD" ) ){
      if( ( lenDSVD + size ) > nDSVD ){
	pDSVD = (char *) realloc( pDSVD, lenDSVD + size );
	nDSVD = lenDSVD + size;
      }
      memcpy( pDSVD+lenDSVD, pData, size );
      lenDSVD += size;
    }

    //
    // DITR 0x0200 *************************************************************
    //
    if( iName == faName2Int( "DITR" ) ){
      int nrow = ArtePointer<DITR>::nrow();
#ifdef _DEB_
      if( FADEB ){
	cout << "  DITR: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DITR>::getLink().getRef(),
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DITR>::getLink().getRef(),
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DITR> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "  DITR: "
	     << "newID=0x"  << hex << *newID   << dec << " "
	     << "nrow()="          << t.nrow()        << " "
	     << "nSparse="         << nSparse         << " "
	     << "red="             << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DITR>::getLink().getRef(), red );
	mmfree_( ArtePointer<DITR>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "  DITR: nrow=" << ArtePointer<DITR>::nrow() << endl;
      }
#endif
    }
    
    //
    // DPTC 0x0300 *************************************************************
    //
    if( iName == faName2Int( "DPTC" ) ){
      int nrow = ArtePointer<DPTC>::nrow();
#ifdef _DEB_
      if( FADEB ){
	cout << "  DPTC: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DPTC>::getLink().getRef(),
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DPTC>::getLink().getRef(),
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DPTC> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "  DPTC: "
	     << "newID=0x"  << hex << *newID   << dec << " "
	     << "nrow()="          << t.nrow()        << " "
	     << "nSparse="         << nSparse         << " "
	     << "red="             << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DPTC>::getLink().getRef(), red );
	mmfree_( ArtePointer<DPTC>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "  DPTC: nrow=" << ArtePointer<DPTC>::nrow() << endl;
      }
#endif
    }

    //
    // DOTR 0x0400 *************************************************************
    //
    if( iName == faName2Int( "DOTR" ) ){
      int nrow = ArtePointer<DOTR>::nrow();
#ifdef _DEB_
      if( FADEB ){
	cout << "  DOTR: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DOTR>::getLink().getRef(),
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DOTR>::getLink().getRef(),
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DOTR> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "  DOTR: "
	     << "newID=0x"  << hex << *newID   << dec << " "
	     << "nrow()="          << t.nrow()        << " "
	     << "nSparse="         << nSparse         << " "
	     << "red="             << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DOTR>::getLink().getRef(), red );
	mmfree_( ArtePointer<DOTR>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "  DOTR: nrow=" << ArtePointer<DOTR>::nrow() << endl;
      }
#endif
    }

    //
    // DRIC 0x0500 *************************************************************
    //
    if( iName == faName2Int( "DRIC" ) ){
      int nrow = ArtePointer<DRIC>::nrow();
#ifdef _DEB_
      if( FADEB ){
	cout << "  DRIC: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DRIC>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DRIC>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DRIC> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "  DRIC: "
	     << "newID=0x" << hex << *newID   << dec << " "
	     << "nrow()="         << t.nrow()        << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DRIC>::getLink().getRef(), red );
	mmfree_( ArtePointer<DRIC>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "  DRIC: nrow=" << ArtePointer<DRIC>::nrow() << endl;
      }
#endif
    }

    //
    // DTRD 0x0600 *************************************************************
    //
    if( iName == faName2Int( "DTRD" ) ){
      int nrow = ArtePointer<DTRD>::nrow();
#ifdef _DEB_
      if( FADEB ){
	cout << "  DTRD: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DTRD>::getLink().getRef(),
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DTRD>::getLink().getRef(),
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DTRD> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "  DTRD: "
	     << "newID=0x"  << hex << *newID   << dec << " "
	     << "nrow()="          << t.nrow()        << " "
	     << "nSparse="         << nSparse         << " "
	     << "red="             << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DTRD>::getLink().getRef(), red );
	mmfree_( ArtePointer<DTRD>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "  DTRD: nrow=" << ArtePointer<DTRD>::nrow() << endl;
      }
#endif
    }

    //
    // DCAL 0x0700 *************************************************************
    //
    if( iName == faName2Int( "DCAL" ) ){
      int nrow = ArtePointer<DCAL>::nrow();
#ifdef _DEB_
      if( FADEB ){
	cout << "  DCAL: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DCAL>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DCAL>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DCAL> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "  DCAL: "
	     << "newID=0x" << hex << *newID   << dec << " "
	     << "nrow()="         << t.nrow()        << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DCAL>::getLink().getRef(), red );
	mmfree_( ArtePointer<DCAL>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "  DCAL: nrow=" << ArtePointer<DCAL>::nrow() << endl;
      }
#endif
    }

    //
    // DMUT 0x0800 *************************************************************
    //
    if( iName == faName2Int( "DMUT" ) ){
      int nrow = ArtePointer<DMUT>::nrow();
#ifdef _DEB_
      if( FADEB ){
	cout << "  DMUT: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DMUT>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DMUT>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DMUT> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "  DMUT: "
	     << "newID=0x" << hex << *newID   << dec << " "
             << "nrow()="         << t.nrow()        << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DMUT>::getLink().getRef(), red );
	mmfree_( ArtePointer<DMUT>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "  DMUT: nrow=" << ArtePointer<DMUT>::nrow() << endl;
      }
#endif
    }

    //
    // DMUX 0x0900 *************************************************************
    //
    if( iName == faName2Int( "DMUX" ) ){
      int nrow = ArtePointer<DMUX>::nrow();
#ifdef _DEB_
      if( FADEB ){
	cout << "  DMUX: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DMUX>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DMUX>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DMUX> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "  DMUX: "
	     << "newID=0x" << hex << *newID   << dec << " "
             << "nrow()="         << t.nrow()        << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DMUX>::getLink().getRef(), red );
	mmfree_( ArtePointer<DMUX>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "  DMUX: nrow=" << ArtePointer<DMUX>::nrow() << endl;
      }
#endif
    }

    //
    // DMUP 0x0a000 *************************************************************
    //
    if( iName == faName2Int( "DMUP" ) ){
      int nrow = ArtePointer<DMUP>::nrow();
#ifdef _DEB_
      if( FADEB ){
	cout << "  DMUP: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DMUP>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DMUP>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DMUP> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "  DMUP: "
	     << "newID=0x" << hex << *newID   << dec << " "
             << "nrow()="         << t.nrow()        << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DMUP>::getLink().getRef(), red );
	mmfree_( ArtePointer<DMUP>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "  DMUP: nrow=" << ArtePointer<DMUP>::nrow() << endl;
      }
#endif
    }

    //
    // DFLT 0x0b00 *************************************************************
    //
    if( iName == faName2Int( "DFLT" ) ){
      int nrow = ArtePointer<DFLT>::nrow();
#ifdef _DEB_
      if( FADEB ){
	cout << "  DFLT: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DFLT>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DFLT>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DFLT> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "  DFLT: "
	     << "newID=0x" << hex << *newID   << dec << " "
             << "nrow()="         << t.nrow()        << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DFLT>::getLink().getRef(), red );
	mmfree_( ArtePointer<DFLT>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "  DFLT: nrow=" << ArtePointer<DFLT>::nrow() << endl;
      }
#endif
    }

    //
    // DSLT 0x0c00 *************************************************************
    //
    if( iName == faName2Int( "DSLT" ) ){
      int nrow = ArtePointer<DSLT>::nrow();
#ifdef _DEB_
      if( FADEB ){
	cout << "  DSLT: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DSLT>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DSLT>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DSLT> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "  DSLT: "
	     << "newID=0x" << hex << *newID   << dec << " "
             << "nrow()="         << t.nrow()        << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DSLT>::getLink().getRef(), red );
	mmfree_( ArtePointer<DSLT>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "  DSLT: nrow=" << ArtePointer<DSLT>::nrow() << endl;
      }
#endif
    }

    //
    // DTLT 0x0d00 *************************************************************
    //
    if( iName == faName2Int( "DTLT" ) ){
      int nrow = ArtePointer<DTLT>::nrow();
#ifdef _DEB_
      if( FADEB ){
	cout << "  DTLT: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DTLT>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DTLT>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DTLT> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "  DTLT: "
	     << "newID=0x" << hex << *newID   << dec << " "
             << "nrow()="         << t.nrow()        << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DTLT>::getLink().getRef(), red );
	mmfree_( ArtePointer<DTLT>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "  DTLT: nrow=" << ArtePointer<DTLT>::nrow() << endl;
      }
#endif
    }

    //
    // DTAR 0x0e00 *************************************************************
    //
    if( iName == faName2Int( "DTAR" ) ){
      int nrow = ArtePointer<DTAR>::nrow();
#ifdef _DEB_
      if( FADEB ){
	cout << "  DTAR: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DTAR>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DTAR>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DTAR> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "  DTAR: "
	     << "newID=0x" << hex << *newID   << dec << " "
             << "nrow()="         << t.nrow()        << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DTAR>::getLink().getRef(), red );
	mmfree_( ArtePointer<DTAR>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "  DTAR: nrow=" << ArtePointer<DTAR>::nrow() << endl;
      }
#endif
    }

    //
    // DFCS 0x0f00 *************************************************************
    //
    if( iName == faName2Int( "DFCS" ) ){
      int nrow = ArtePointer<DFCS>::nrow();
#ifdef _DEB_
      if( FADEB ){
	cout << "  DFCS: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DFCS>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DFCS>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DFCS> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "  DFCS: "
	     << "newID=0x" << hex << *newID   << dec << " "
             << "nrow()="         << t.nrow()        << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DFCS>::getLink().getRef(), red );
	mmfree_( ArtePointer<DFCS>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "  DFCS: nrow=" << ArtePointer<DFCS>::nrow() << endl;
      }
#endif
    }

    //
    // DDAQ
    //
    if( iName == faName2Int( "DDAQ" ) ){
      ArtePointer<DDAQ> t( 2 );
      // DAQ data structure
      EVT_DATA *p = (EVT_DATA *) &t->bits;

      //
      // fill EVHD from DDAQ
      //

      // run number
      ESP_get_int( &p->run_number, &evhd->run, 1 );
      arteflag_.numrun = evhd->run;
      // exp number
      ESP_get_int( &p->exp_number, &evhd->exp, 1 );
      arteflag_.numexp = evhd->exp;
#ifdef _DEB_
      if( FADEB ){
	cout << "  DDAQ: "
	     << "evhd->evt=" << evhd->evt << " "
	     << "evhd->run=" << evhd->run << " "
	     << "evhd->exp=" << evhd->exp << endl;
      }
#endif

      //
      // fill DIHD from DDAQ
      //
      short tmp = 0;
      ESP_get_short( &p->physBX, &tmp,  1 );
      dihd->bx  = int( tmp );
      ESP_get_short( &p->fltevt, &tmp,  1 );
      dihd->flt = int( tmp );
      dihd->slb = slbHdr.slbid;
#ifdef _DEB_
      if( FADEB ){
	cout << "  DIHD: "
	     << "dihd->bx=0x"  << hex << dihd->bx         << " "
	     << "dihd->flt=0x"        << dihd->flt        << " "
	     << "dihd->slb=0x"        << dihd->slb << dec << endl;
      }
#endif
    }
  }

  //
  // print run, event
  //
  if( FADEB ){ 
    faPrintEvtCount( faEVHDrun(), faEVHDevent() );
  }
  
  //
  // build DSVD
  //
  if( lenDSVD ){
    int recID = 0x2000;
    mmbook_( ArtePointer<DSVD>::getLink().getRef(),
	     lenDSVD / sizeof( int ) + 1 );
    ArtePointer<DSVD> t( 1 );
    int  *newID = (int *)   &t->bits;    
    char *p     = (char *)( &t->bits + 1 );
    // sparsify?
    int nSparse = 0;
    if( !NOSPAR ){
      nSparse = faCopySlb( &recID, pDSVD, newID, p, lenDSVD );
    }
    else {
      nSparse = faCopySlb( &recID, pDSVD, newID, p, lenDSVD );
    }
    int red = ( nSparse - lenDSVD ) / int( sizeof( int ) );
#ifdef _DEB_
    if( FADEB ){
      cout << "  DSVD: "
	   << "newID=0x" << hex << *newID   << dec << " "
	   << "nrow()="         << t.nrow()        << " "
	   << "nSparse="        << nSparse         << " "
	   << "red="            << red             << endl;
    }
#endif
    if( red ){
      mmbook_( ArtePointer<DSVD>::getLink().getRef(), red );
      mmfree_( ArtePointer<DSVD>::getLink().getRef() );
    }
  }

  // free memory
  free( pDSVD );

  return ARTE_GOOD;
}
