//--------------------------------------------------------------------
// Created: 02.03.98 by   Olya Igonkina
/* MARPLE function for matching SVD segments and ECAL clusters

   FORTRAN_ROUTINE(matchsvdecal)( float * threshold_cut )
   if threshold_cut > 200 - it is Ktrig cut, otherwise it is pt_cut  

   1. search of pair segment-cluster corresponded to electrons/positrons 
      hypothesis is performed. Two algorithms are available
      a) segment momentum estimated by its prolongation 
         to the cluster point coordinate.
      b) momentum is supposed to be equal a energy of the cluster with  
         2 charge hypothesis
	 
   2. search of bremstrahlung is performed for segments satisfied charge 
      hypothesis ( linear prolongation of segment to the ECAL frontplate, 
      chi2 value is calculated). The energy of bremstrahlung cluster is added
      to charge hypothesis's one.

 */
//--------------------------------------------------------------------

#include "marple/marple.hh"
#include "marple/RcSegClusBuilder.hh"
#include "arte/RSEG_cmp.hh"

#include "marple/RcMcFunction.hh"
// #include "PartFunction.hh"

//definition of constants for SUSI-CARE matching ( see \Ref{RcMatchConst})
// flag, z_plane, z_min, z_max, use_pair_cut, n_pair_left, n_pair_right, 
//use_delta_cut, delta_x, delta_y, delta_tx, delta_ty, delta_pe, chi2

// match with Ecluster
RcMatchConst SCmatchEConst = { 
  11000,     1330.,   0.,  1500.,      1,          2,            2,
  1,       8.,      5.,   1.e+10,  1.e+10,   200.,   10.
};


// match with P estimation
RcMatchConst SCmatchPConst = { 
 11001,     1330.,   0.,  1500.,      1,          2,            2,
   1,       1.,      4.,   1.e+10,  1.e+10,   0.04,   35.
};

// constant for bremstrahlung search
RcMatchConst BremConst = { 
 11000,     1330.,   0.,  1500.,      1,          1,            1,
     1,       6.,      6.,   1.e+10,  1.e+10,   1.e+10,   6.
};
RcMatchConst BremConstLow = { 
 11000,     1330.,   0.,  1500.,      1,          1,            1,
     1,       5.,      5.,   1.e+10,  1.e+10,   1.e+10,   10.
};


extern "C"  
void FORTRAN_ROUTINE(matchsvdecal)( float * threshold_cut ) 
{

  // ----------------------------------------------------
  //  if magnet is OFF call more trivial algorithm!
  if( !maMagnetON() ) 
    {      
      matchsenomag_( threshold_cut);
      return;
    }

  // ----------------------------------------------------

  RcSegClusBuilder gISvdEcalMatch;  // builder for charge track search
  RcSegClusBuilder gISvdEcalBrem;   // builder for bremstrahlung search


  //--------------------------------------------------------------------
  // put constants to the MatchBuilder
  //--------------------------------------------------------------------
  gISvdEcalMatch.setConst(     SCmatchEConst);
  gISvdEcalMatch.setBremConst( BremConstLow);
  
  // work with low energetic clusters for bremstrahlung search
  gISvdEcalBrem.setConst(     SCmatchEConst);
  gISvdEcalBrem.setBremConst( BremConstLow);
  
  //--------------------------------------------------------------------
  // create vector of clusters from CARE
  //--------------------------------------------------------------------

  int iErr  = 0;
  for( ArteTableIterator <RCCL> irccl; 
       irccl.valid(); irccl.next() )
    {
      irccl->flag = -1;
      int takeRCCL = 0;
      if( *threshold_cut > 200 )
	{
	  // Ktrig cut
	  HepDouble radius = sqrt(irccl->x*irccl->x + irccl->y*irccl->y);
	  HepDouble radMag = sqrt(irccl->x*irccl->x + 
				  fabs(irccl->y*irccl->y*irccl->y) );
	  if( radMag && radius && 
	      irccl->esum > *threshold_cut * ( 1/radius+1/radMag ) )
	    takeRCCL = 1;
	}else{
	  // pt_cut
	  HepDouble radius = sqrt(irccl->x*irccl->x + irccl->y*irccl->y);
	  HepDouble Et = irccl->esum * radius/
	    sqrt( radius*radius + irccl->z*irccl->z );
	  if( Et > *threshold_cut ) takeRCCL = 1;
	}
      RcClusPoint * ptPoint = new RcClusPoint( irccl);
      if( takeRCCL)
	{
	  gISvdEcalMatch.addPointRight( *ptPoint);
	}else  {
	  gISvdEcalBrem.addPointRight( *ptPoint);
	}
      ptPoint = 0;
    }  

  //--------------------------------------------------------------------
  // create vector of segments from SUSI
  //--------------------------------------------------------------------
  
  for( ArteTableIterator <RSEG> irseg;
       irseg.valid(); irseg.next() )
    {
      //if( IrcSEmatchPConst.geomAccept(double(irseg->ze)) )
      if( irseg->fit >= 0 && (irseg->cmp & Rsegc::vxd)  && 
	  gISvdEcalMatch.ptToMatchConst()->geomAccept(HepDouble(irseg->ze)) )
	{ 	  
	  RcSegPoint * ptToPoint = new RcSegPoint( irseg , 1); // endPoint
	  gISvdEcalMatch.addPointLeft( *ptToPoint);
	  ptToPoint = 0;
	}
    }
  //--------------------------------------------------------------------
  // do Matching 
  //--------------------------------------------------------------------

  //--------------------------------------------------------------------
  // I step : search for electron/positron candidates
  //--------------------------------------------------------------------

  //gISvdEcalMatch.matchWithMomentum();
  gISvdEcalMatch.matchWithEcluster();


  //--------------------------------------------------------------------
  // Ia step : analysis of charge pair, candidates for bremstrahlung
  //--------------------------------------------------------------------

  unsigned int charge = gISvdEcalMatch.matchList().size();
  if( charge == 0 )
    {
      // no charged pair could be found
      //gISvdEcalMatch.stat(2);
      gISvdEcalMatch.deleteAll();  
      gISvdEcalBrem.deleteAll();  
      cout << " no charge candidate is found " << endl;
      return ;
    }
  
  // interesting entry for bremstrahlung search
  vector< RcSegPoint *> vecTmp = gISvdEcalMatch.vectorLeft(); 
  for( unsigned int iMatch = 0;  iMatch<charge; iMatch++)
    {
      for( unsigned int iBremSi = 0;  iBremSi < vecTmp.size(); iBremSi++ )
	{
	  int found = 0;
	  if(  gISvdEcalMatch.leftOrigin(iMatch)->link() == 
	       vecTmp[iBremSi]->link() ) found = 1;
	  if( found ) 
	    {
	      RcSegPoint * ptToPoint = new 
		RcSegPoint(      * gISvdEcalMatch.leftOrigin(iMatch));
	      gISvdEcalBrem.addPointLeft( *ptToPoint );
	      ptToPoint = 0;
	    }
	}
    }
  
  //cout << " number of electrons" << electronVec.size() 
  //   << " " << positronVec.size() << endl;
  
  //--------------------------------------------------------------------
  // II step : search for bremstralung
  //--------------------------------------------------------------------
  
  // search in the high energetic clusters set
  //gISvdEcalMatch.matchBrem();
  
  // search in the low energetic clusters set
  //gISvdEcalBrem.matchBrem();
  
  //--------------------------------------------------------------------
  // outprint
  //--------------------------------------------------------------------

  //gISvdEcalMatch.fillMatch(1000);
  //gISvdEcalMatch.fillDeltas(2000, 3000);

  //gISvdEcalBrem.fillMatch(1000);
  //gISvdEcalBrem.fillDeltas(2000);

  /* 
  static int firstSVDCARE = 1;
  if( firstSVDCARE )
    {
      cout << "================= match ============="<< endl; 
      gISvdEcalMatch.stat(1);
      cout << endl << endl << endl;
      cout <<" ================= Bremstralung ============="<< endl; 
      gISvdEcalBrem.stat(0);
      firstSVDCARE = 0;
    }
  
  cout << "================= match ============="<< endl; 
  gISvdEcalMatch.stat(0,0);
  cout <<" ================= Bremstralung ============="<< endl; 
  gISvdEcalBrem.stat(0,-1);
  */
  
  //--------------------------------------------------------------------
  // save found relation to the table
  //--------------------------------------------------------------------

  gISvdEcalMatch.saveRelation(iErr);
  if( iErr ){ cout << "MARPLE: some relations were not saved " << endl;}
  gISvdEcalBrem.saveRelation(iErr);
  if( iErr ){ cout << "MARPLE: some relations were not saved 1 " << endl;}

  //--------------------------------------------------------------------
  // clean everything
  //--------------------------------------------------------------------
  gISvdEcalMatch.deleteAll();  
  gISvdEcalBrem.deleteAll();  

 
}
