/* This file defines SetupMenu<AnySetup> class.
   It contains 
     IstreamMenu_1Par<AnySetup*>          setup;
     IstreamMenu_1Par<AnySetup::IChamber> chamber;
     IstreamMenu_1Par<AnySetup::ISection> section;
     IstreamMenu_1Par<AnySetup::ICell>    cell;
   E.g. to process the top menu "setup" call: setup(is, AnySetup*);
   All next menu's are attached to previous in such a way that
   in e.g. "section" menu you can deal only with sections from the 
   chambers and setups selected in the "setup" and "chamber" menu's. 

   To insert new interactive command with one parameter in the menu
   you may write (see file interactive_1par.hh) e.g.:
     section("command_name") = make_interactive_command_1par(
     <your_function_object>, (AnySetup::ISection*)0);
   Here <your_function_object> should be able to work with 
   istream* is and AnySetup::ISection i_section:
     expression <your_function_object>(is, i_section) should be valid.
*/   

// Author V.Balagura (balagura@mail.desy.de)   (28.02.99).

#ifndef _setup_menu_hh
#define _setup_menu_hh 1

#include "ptset/istream_menu_1par.hh"
#include "ptset/call_via_istr.hh"
#include "ptset/string_primitives.hh"
#include "ptset/setup.hh"
#include "ptset/setup_figure_istream_menu.hh"

/* ----------------------- SetupMenuPrimitives ----------------------- */
template <class AnySetup>
class SetupMenuPrimitives {
 public:
  /* -------------------- ChamberMenu -------------------- */
  class ChamberMenu {
    IstreamMenu_1Par<typename AnySetup::IChamber>& menu;
  public:
    ChamberMenu(IstreamMenu_1Par<typename AnySetup::IChamber>& m) 
      : menu(m) {}
    void operator()(istream* is, AnySetup* p_setup) {
      string chamber_name;
      *is >> chamber_name;
      if ( !is ) {
	global_error()("can not read chamber name");
	return;
      }
      typename AnySetup::IChamber i_chamber = p_setup->find(chamber_name);
      if ( i_chamber == p_setup->end() ) {
	global_error()("there is no chamber \"" + chamber_name + "\"");
	return;
      }
      menu(is, i_chamber);
      if ( global_error() )
	global_error() << ("chamber \"" + chamber_name + "\"");
    }
  };
  /* -------------------- SectionMenu -------------------- */
  class SectionMenu {
    IstreamMenu_1Par<typename AnySetup::ISection>& menu;
  public:
    SectionMenu(IstreamMenu_1Par<typename AnySetup::ISection>& m) 
      : menu(m) {}
    void operator()(istream* is, typename AnySetup::IChamber i_chamber) {
      unsigned int n_section;
      *is >> n_section;
      if ( !is ) {
	global_error()("can not read section number");
	return;
      }
      if ( n_section >= (*i_chamber).second.size() ) {
	string tmp("there is no section #");
	tmp << n_section;
	global_error()(tmp);
	return;
      }
      typename AnySetup::ISection 
	i_section = (*i_chamber).second.begin() + n_section;
      menu(is, i_section);
      if ( global_error() ) {
	string tmp("section #");
	tmp << n_section;
	global_error() << tmp;
      }
    }
  };
  /* -------------------- CellMenu -------------------- */
  class CellMenu {
    IstreamMenu_1Par<typename AnySetup::ICell>& menu;
  public:
    CellMenu(IstreamMenu_1Par<typename AnySetup::ICell>& m) 
      : menu(m) {}
    void operator()(istream* is, typename AnySetup::ISection i_section) {
      Index ind;
      *is >> ind;
      if ( !is ) {
	global_error()("can not read cell index");
	return;
      }
      typename AnySetup::ICell i_cell = (*i_section).find(ind);
      if ( i_cell == (*i_section).end() ) {
	string tmp("there is no cell with index ");
	tmp << ind;
	global_error()(tmp);
	return;
      }
      menu(is, i_cell);
      if ( global_error() ) {
	string tmp("cell with index ");
	tmp << ind;
	global_error()(tmp);
	return;
      }
    }
  };
  /* -------------------- Functions -------------------- */
  static void create_chamber(const string& chamber_name, AnySetup* p_setup) {
    AnySetup& s = *p_setup;
    s.erase(chamber_name);
    s[chamber_name];
  }
  static void create_sections(unsigned int n_sections,
			      typename AnySetup::IChamber i_chamber) {
    for (unsigned int i=0; i<n_sections; ++i)
      (*i_chamber).second.push_back(typename AnySetup::Section());
  }
  static void erase_section(unsigned int n_section,
			    typename AnySetup::IChamber i_chamber) {
    typename AnySetup::Chamber& ch = (*i_chamber).second;
    if ( n_section >= ch.size() ) {
      string tmp("there is no section #");
      tmp << n_section;
      global_error()(tmp);
      return;
    }
    ch.erase(ch.begin()+n_section);
  }
  static void set_chamber_z(double z,
			    typename AnySetup::IChamber i_chamber) {
    (*i_chamber).second.z = z;
  }
  static void set_section_grid(const Grid& grid, 
			       typename AnySetup::ISection i_section) {
    (*i_section).grid = grid;
  }
  static void set_section_figure(istream* is,
				 typename AnySetup::ISection i_section) {
    SetupFigureBase* fig;
    setup_figure_istream_menu()(is, &fig);
    if ( !global_error() ) (*i_section).ptr_figure = fig;
  }
  static void make_chamber_cells(const Rectangle& cover, 
				 typename AnySetup::IChamber i_chamber) {
    (*i_chamber).second.make_cells(cover);
  }
  static void make_setup_cells_in_projective_rectangle
  (const Rectangle& cover_at_first_chamber, AnySetup* p_setup) {
    // find the minimal z (of the chamber with the lowest z)
    vector<vector<typename AnySetup::IChamber> > vv_ichamber;
    // z_precision = 0.1
    sort_setup_chambers_by_z(*p_setup, &vv_ichamber, 0.1);
    double z0 = (*vv_ichamber[0][0]).second.z;

    for (typename AnySetup::IChamber ich=p_setup->begin(); 
	 ich!=p_setup->end(); ++ich) {
      typename AnySetup::Chamber& ch = (*ich).second;
      ch.make_cells(cover_at_first_chamber * (ch.z/z0));
    }
  }
  static void dump_setup_structure(AnySetup* p_setup) {
    p_setup->dump_structure(cout);
  }
};

/* ------------------------- SetupMenu ------------------------- */
template <class AnySetup>
class SetupMenu {
 public:
  IstreamMenu_1Par<AnySetup*>                   setup;
  IstreamMenu_1Par<typename AnySetup::IChamber> chamber;
  IstreamMenu_1Par<typename AnySetup::ISection> section;
  IstreamMenu_1Par<typename AnySetup::ICell>    cell;

  SetupMenu() {
    
    setup.parser   = new SlashBlock_InteractiveParser;
    chamber.parser = new SlashBlock_InteractiveParser;
    section.parser = new SlashBlock_InteractiveParser;
    cell.parser    = new SlashBlock_InteractiveParser;

    setup("chamber") = make_interactive_command_1par
     (SetupMenuPrimitives<AnySetup>::ChamberMenu(chamber), (AnySetup**)0);
    chamber("section") = make_interactive_command_1par
      (SetupMenuPrimitives<AnySetup>::SectionMenu(section),
       (typename AnySetup::IChamber*)0);
    section("cell") = make_interactive_command_1par
      (SetupMenuPrimitives<AnySetup>::CellMenu(cell),
       (typename AnySetup::ISection*)0);

    setup("create") = make_interactive_command_1par
      (call_via_istr_check_EOF_1par
       (&SetupMenuPrimitives<AnySetup>::create_chamber,
	(string*)0, (AnySetup**)0), (AnySetup**)0);
    chamber("create") = make_interactive_command_1par
      (call_via_istr_check_EOF_1par
       (&SetupMenuPrimitives<AnySetup>::create_sections,
	(unsigned int*)0, (typename AnySetup::IChamber*)0),
       (typename AnySetup::IChamber*)0);
    chamber("erase") = make_interactive_command_1par
      (call_via_istr_check_EOF_1par
       (&SetupMenuPrimitives<AnySetup>::erase_section,
	(unsigned int*)0, (typename AnySetup::IChamber*)0),
       (typename AnySetup::IChamber*)0);
    chamber("z") = make_interactive_command_1par
      (call_via_istr_check_EOF_1par
       (&SetupMenuPrimitives<AnySetup>::set_chamber_z,
	(double*)0, (typename AnySetup::IChamber*)0), 
       (typename AnySetup::IChamber*)0);
    section("grid") = make_interactive_command_1par
      (call_via_istr_check_EOF_1par
       (&SetupMenuPrimitives<AnySetup>::set_section_grid,
	(Grid*)0, (typename AnySetup::ISection*)0),
       (typename AnySetup::ISection*)0);
    section("figure") = make_interactive_command_1par
      (&SetupMenuPrimitives<AnySetup>::set_section_figure,
       (typename AnySetup::ISection*)0);
    chamber("make_cells") = make_interactive_command_1par
      (call_via_istr_check_EOF_1par
       (&SetupMenuPrimitives<AnySetup>::make_chamber_cells,
	(Rectangle*)0, (typename AnySetup::IChamber*)0),
       (typename AnySetup::IChamber*)0);
    setup("make_cells_in_projective_rectangle") = 
     make_interactive_command_1par
     (call_via_istr_check_EOF_1par
      (&
       SetupMenuPrimitives<AnySetup>::make_setup_cells_in_projective_rectangle,
	    (Rectangle*)0, (AnySetup**)0), (AnySetup**)0);
    setup("dump_structure") = make_interactive_command_1par
    (call_via_istr_check_EOF_1par
     (&SetupMenuPrimitives<AnySetup>::dump_setup_structure,
      (AnySetup**)0), (AnySetup**)0);
  }

  SetupMenu(const SetupMenu& m) { assign(m); }

  SetupMenu& operator=(const SetupMenu& m) {
    if ( &m == this ) return *this;
    assign(m);
    return *this;
  }
private:
  void assign(const SetupMenu& menu) {
    setup   = menu.setup;
    chamber = menu.chamber;
    section = menu.section;
    cell    = menu.cell;
    // now make relations
    setup("chamber") = make_interactive_command_1par
     (SetupMenuPrimitives<AnySetup>::ChamberMenu(chamber), (AnySetup**)0);
    chamber("section") = make_interactive_command_1par
      (SetupMenuPrimitives<AnySetup>::SectionMenu(section),
       (typename AnySetup::IChamber*)0);
    section("cell") = make_interactive_command_1par
      (SetupMenuPrimitives<AnySetup>::CellMenu(cell),
       (typename AnySetup::ISection*)0);
  }
};

#endif
