// 
//  Author:  Rainer Mankel                     Created:  29-Sep-98
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// 
#include <iostream.h>
#include <math.h>
#include "arte/arte.hh"
#include "arte/ArteMessage.hh"
#include "rang/State.hh"
#include "rang/Node.hh"
#include "rang/HitBase.hh"
#include "rang/ScatThick.hh"
#include "rang/Geom.hh"
#include "rang/zfpar.hh"
#include "rang/utfeld.hh"
#include "rang/RadEl.hh"


//-- Constructor
rangerHitBase::rangerHitBase() {
  cout << "This is the rangerHitBase default constructor" << endl;
  z = 0;
  m = 0;
  V = 0;
  H = 0;
  r = 0;
  R = 0;
  chiSq = 0;
};

rangerHitBase::rangerHitBase(ArtePointer<RHIT> rhit, rangerFitMode fm) 
  : rangerNode(rhit,fm) {
#ifdef FULL
  cout << "This is the rangerHitBase(rhit) constructor, hitb=" 
       << rhit->hitb << endl;
#endif
  m = 0;
  V = 0;
  H = 0;
  r = 0;
  R = 0;
  chiSq = 0;
  z = 0;
};

rangerHitBase::rangerHitBase(ArtePointer<HITB> hitb, rangerFitMode fm) 
  : rangerNode(hitb,fm) {
#ifdef FULL
  cout << "This is the rangerHitBase(hitb) constructor, hitb=" 
       << hitb << endl;
#endif
  m = 0;
  V = 0;
  H = 0;
  r = 0;
  R = 0;
  chiSq = 0;
  z = 0;
};

rangerHitBase::rangerHitBase(ArtePointer<RCCL> rccl,int charge,rangerFitMode fm) 
  : rangerNode(rccl,charge,fm) {
#ifdef FULL
  cout << "This is the rangerHitBase(rccl) constructor, rccl=" 
       << rccl << endl;
#endif
  m = 0;
  V = 0;
  H = 0;
  r = 0;
  R = 0;
  chiSq = 0;
  z = rccl->z;
};

rangerHitBase::~rangerHitBase(){
  delete H;
  delete r;
  delete R;
};

void rangerHitBase::processNoise(rangerParticleType pType) {

  rangerScatThin scat( ms(), state->qPEst(),
		       state->tx(),
		       state->ty());

  state->applyScat( scat );

  if (pType == ELECTRON) {
    rangerRadEl rad( radiationThickness(), predictedState->qP() );
    predictedState->applyRadEl( rad );
  }

  state->applyDeDx( radiationThickness() );

}



void rangerHitBase::predProcessNoise(rangerParticleType pType) {

  rangerScatThick scat( ms(), predictedState->qPEst(),
			predictedState->tx(),
			predictedState->ty(), -dz() );

  predictedState->applyScat( scat );
  
  if (pType == ELECTRON) {
    rangerRadEl rad( radiationThickness(), predictedState->qP() );
    predictedState->applyRadEl( rad );
  }

  predictedState->applyDeDx( radiationThickEff() );

#ifdef FULL
  cout << "rangerHitBase::processNoise: correction applied, cov matrix is now "
       << predictedState->cov() << endl;
#endif
}


int rangerHitBase::filter() { 

#ifdef FULL
  cout << "rangerHitBase::filter(): this is the node at z=" << z << endl;
  list();
#endif
  if (! predictedState) {
    predictedState = rangerState::makeState( state->fitMode() );
  }
  else {
  }

  *(predictedState) = *state;      // copy initialization into prediction
  predictedState->covInit();      // initialize cov matrix
                                  // no transport matrix needed
  int absorbStatus = absorb();
#ifdef FULL
  cout << "rangerHitBase::filter(): absorb returns status" 
       << absorbStatus << endl;
#endif
  return absorbStatus;
}






int rangerHitBase::filter(rangerNode* n,rangerParticleType pType) {

// for the transport matrix, the predicted state & the projection matrix
// allocation is done dynamically: 
//   if they do not exist, they are created,
//   if they do exist, they are reused.
// on destruction of the Node, both are destroyed

#ifdef FULL
  cout << "rangerHitBase::filter(...): this is the node at z=" << z << endl;
  cout << "                         call prediction" << endl;
  if (predictedState) cout << "      predictedState OK";
  else {
    cout << "      predictedState " << predictedState;
  }
  if (transportMatrix) cout << "     transportMatrix OK " << endl;
  else {
    cout << "    transportMatrix " << transportMatrix << endl;
  }
  list();
#endif

  int returnStatus = n->predict(z,predictedState,transportMatrix);

  if (returnStatus < 0) {
    cout << "rangerHitBase::filter: predict returns returnStatus=" << 
      returnStatus << endl;
    return returnStatus;
  }

#ifdef FULL
  cout << "rangerHitBase::filter(...): the predicted state is" 
       << predictedState->vec() << endl
       << predictedState->cov() << endl
       << "                         the transport matrix is "
       << *transportMatrix << endl;
#endif
  
  if (plane() != n->plane())    predProcessNoise(pType);
  int absorbStatus = absorb();
#ifdef FULL
  cout << "rangerHitBase::filter(...): the filtered state is"
       << state->vec() << endl;
  list();
  if (state->tx() >0 || state->tx() <=0);
  else {
    cout << "rangerHitBase::filter(...): invalid tx parameter" << endl;
    exit(1);
  }
#endif

  return absorbStatus;
}
  


int rangerHitBase::absorb() { 

  if (! H) {
    H = new HepMatrix(dim(),predictedState->dim());
    if (! H) {
      cout << "rangerHitBase::absorb(): matrix allocation failed, H=" << H 
	   << endl;
      exit(1);
    }
  }

  if (!r)  r = new HepVector( dim() );   // residual

  predictedState->project(*r,(*H),this);
#ifdef FULL
  cout << "HitBase::absorb: back from project, matrix=" << endl;
  cout << *H << endl;
  cout << "The residual is " << *r << endl;
#endif

  int invertStatus = 0;

#ifdef FULL
  cout << "HitBase::absorb: start calculation of K matrix" << endl;
  cout << "The predicted state cov matrix is " << predictedState->cov()
       << endl;
  cout << "The H matrix is " << (*H) << endl;
  cout << "The V matrix is " << (*V) << endl;
#endif

  HepMatrix K( predictedState->cov() * H->T() 
       * ((*V) + predictedState->cov().similarity(*H)).inverse(invertStatus) );

#ifdef FULL
  cout << "K matrix after assignment " << K << endl;
#endif

#ifdef FULL
  if (dim() == 2) {
    cout << "<any key...>" << endl;
    char cc = cin.get();
  }
#endif

  if (invertStatus != 0) {
    arteMsg << "HitBase::absorb: invert status " << invertStatus << endl;
    static int id1;
    arteMsg.allp(id1,1);
    
    return -10;
  }

  state->vec() = predictedState->vec() + K * (*r);

  if ( fabs(state->x()) < 700 &&   // check validity of rkmain input
       fabs(state->y()) < 700 &&
       fabs(state->tx()) < 10   &&
       fabs(state->ty()) < 10   &&
       fabs(state->qP()) < 100 ) ;
  else {
#ifdef FULLSECOND
    cout << "HitBase::absorb: invalid state="
	 << state->vec() 
	 << " at z=" << state->z() << endl;
    if (predictedState) {
      cout << "The prediction was state=" << predictedState->vec()
	   << " cov=" << predictedState->cov() << endl;
    }
    else {
      cout << "There is no predicted state!" << endl;
    }
    //char cc = cin.get();
#endif
    arteMsg << "HitBase::absorb: second kind error";
    static int id2;
    arteMsg.allp(id2,1);
    
    return -10;
  }









#ifdef FULL
  cout << " state->vec after assignment =" << state->vec() << endl;
#endif

  HepMatrix unit(predictedState->dim(),predictedState->dim(),1);

  state->cov().assign( (unit - K * (*H)) * predictedState->cov() );

  HepMatrix unitM(dim(),dim(),1);

  (*r) = (unitM - (*H) * K) * (*r);

  if (!R) R = new HepSymMatrix( dim() );
  (*R).assign( (unitM - (*H) * K) * (*V) );

  invertStatus = 0;
  double ch2 =  (r->T() 
    * R->inverse(invertStatus) * (*r))(1,1);
  if (invertStatus >= 0) {
    chiSq = (ch2>0) ? ch2 : 0;
  }
  else {
    cout << "HitBase::absorb: invert error in chiSq calculation " << endl;
    chiSq = 0;
  }

#ifdef FULL
  cout << "Filtering done" << endl << endl << endl << endl;
#endif
  
  return 0;
}

void rangerHitBase::residUpdate() {
  // update the residual
  HepMatrix HNew(dim(),predictedState->dim());
  state->project(*r,HNew,this);
}


void rangerHitBase::residUpdSmooth(rangerNode* n,HepMatrix& A) {
  // update the residual with smoothing
  HepMatrix HNew(dim(),predictedState->dim());
  state->project(*r,HNew,this);

  (*R) = (*R) - 
    ( (n->getState().cov() - n->getPrediction()->cov()).similarity(A) ).similarity(*H);
  if ( (*R)(1,1) <= 0) {
    cout << " rangerHitBase::residUpdate: R(1,1)=" << (*R)(1,1) << endl;
    chiSq = 0;
  }
  else {
    int invertStatus = 0;
    double ch2 =  (r->T() 
		   * R->inverse(invertStatus) * (*r))(1,1);
    if (invertStatus >= 0) {
      chiSq = (ch2>0) ? ch2 : 0;
    }
    else {
      cout << " rangerHitBase::residUpdate: R invert failed" << endl;
      exit(1);
    }
  }
}


