// 
//  Author:  Rainer Mankel                     Created:  24-Aug-98
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// 

#undef TRANSCD

#include <iostream.h>
#include <math.h>
#include "CLHEP/Matrix/Matrix.h"
#include "CLHEP/Matrix/SymMatrix.h"
#include "rang/StateP.hh"
#include "rang/zfpar.hh"
#include "rang/rkmain.hh"
#include "rang/utfeld.hh"
#include "rang/FitMode.hh"
#include "rang/RadEl.hh"
#include "rang/rktranscd.hh"

//-- Constructor
rangerStateP::rangerStateP() : p(5), c(5) {
//  cout << "rangerState default constructor" << endl;
  zState = 0;
  qPEstimate = 0.1;
#ifdef FULL
  cout << "rangerState rhit constructor" << endl;
  cout << "rangerState: the state vector is now " << endl;
  cout << p << endl;
#endif
}

rangerStateP::rangerStateP(ArtePointer<RHIT> rhit) 
  : rangerState(rhit), p(5), c(5) {
  setX() = rhit->x;
  setY() = rhit->y;
  setTx() = rhit->tx;
  setTy() = rhit->ty;
  setQp() = rhit->p;
}

rangerStateP::rangerStateP(ArtePointer<HITB> hitb) 
  : rangerState(hitb), p(5), c(5) {
  setX() = 0;
  setY() = 0;
  setTx() = 0;
  setTy() = 0;
  setQp() = 0;
}

rangerStateP::rangerStateP(ArtePointer<RCCL> rccl,int charge) 
  : rangerState(rccl), p(5), c(5) {
  setX() = rccl->x;
  setY() = rccl->y;
  setTx() = rccl->tx;
  setTy() = rccl->ty;
  if (rccl->esum > 0) {
    setQp() = charge/rccl->esum;
  }
  else {
    setQp() = charge * 9999;
  }
}

rangerStateP::rangerStateP(const rangerStateP& s) 
  : rangerState(s), p(5), c(5) {
#ifdef FULL
  cout << "rangerStateP::rangerStateP: here is the copy constructor" << endl;
#endif
  zState = s.zState;
  p = s.p;
  c = s.c;
  qPEstimate = s.qPEstimate;
#ifdef FULL
  cout << "rangerStateP::rangerStateP: Copy constructor done " << endl;
#endif
}

rangerStateP::~rangerStateP(){
//  cout << "rangerStateP destructor has been called" << endl;
}

ostream& operator<<(ostream &os,const rangerStateP &s){
  os << s.vec();
  return os;
}

rangerState& rangerStateP::operator=(const rangerStateP& s) {
#ifdef FULL
  cout << "rangerState& rangerStateP::operator=(const rangerStateP& s): "
       << " here is the assignment operator" << endl;
  double zz = s.z();
  cout << "   the given z is " << zz << endl;
#endif
  zState = s.zState;
  qPEstimate = s.qPEst();
  p = s.p;
  c = s.c;
  return *this;
}

rangerState& rangerStateP::operator=(const rangerState& s) {
#ifdef FULL
  cout << "rangerState& rangerStateP::operator=(const rangerState& s)" << endl;
#endif
  zState = s.z();
  qPEstimate = s.qPEst();
  p = s.vec();
  c = s.cov();
  return *this;
}


HepVector& rangerStateP::vec() { return p; }
const HepVector& rangerStateP::vec() const { return p; }
HepSymMatrix & rangerStateP::cov() { return c; } 
const HepSymMatrix & rangerStateP::cov() const { return c; } 

void rangerStateP::covInit() {
#ifdef FULL
  cout << "rangerStateP::covInit: z=" << z << endl;
  cout << "                     state matrix has " 
       << c.num_row() << " " << c.num_col() << endl;
#endif
  c(1,1) = zfpar_.cxxrf;
  c(2,2) = zfpar_.cyyrf;
  c(3,3) = zfpar_.ctxtxrf;
  c(4,4) = zfpar_.ctytyrf;
  c(5,5) = zfpar_.cqqrf;
  for (int i=1; i<=5; i++) {
    for (int j=1; j<i; j++) {
//      cout << "initialize c(" << i << "," << j << ")" << endl;
      c(i,j) = 0;
    }
  }
#ifdef FULL
  cout << "cov matrix is " << endl << c << endl;
#endif
}



int rangerStateP::predict(const double& zDestination,
			 const double& qPExp,rangerState*& s,
			 HepMatrix*& F) {

  int newState     = (s == 0);  // denote if memory allocations exist
  int newTransport = (F == 0);   //   .    .    .        .         .

#ifdef FULL
  cout << "rangerStateP::predict: zDestination=" << zDestination << endl;

  cout << "rangerStateP::predict: targetState=" << s
       << " newState= " <<  newState << endl;
#endif

  if (newState) {
    s = new rangerStateP();
    if (!s) {
      cout << "rangerState::predict: allocation error" << endl; 
      exit(1);
    }
  }
  if (! F) {
    F = new HepMatrix(5,5);
    if (!F) {
      cout << "rangerState::predict: allocation error" << endl; 
      exit(1);
    }
  }

#ifdef FULL
  cout << "rangerStateP::predict: after allocation: s->vec()=" << s->vec() 
       << endl << "       *F=" << *F << endl;
#endif

  double deltaZ = fabs(zDestination - zState);

  if ( (zDestination>zfpar_.zmaxMag && zState>zfpar_.zmaxMag)
       || (zDestination<zfpar_.zminMag && zState<zfpar_.zminMag)
       || deltaZ < 0.001) {
    s->setZ() = zDestination;      // linear transport
    s->setX() = x() + (zDestination - zState) * tx();
    s->setY() = y() + (zDestination - zState) * ty();
    s->setTx() = tx();
    s->setTy() = ty();
    s->setQp() = qP();
    s->setQpEst() = qPEst();

    for (int i=1; i<=5; i++) {
      for (int j=1; j<=5; j++) {
	(*F)(i,j) = 0.;
      }
    }
    (*F)(1,1) = 1;
    (*F)(2,2) = 1;
    (*F)(3,3) = 1;
    (*F)(4,4) = 1;
    (*F)(5,5) = 1;
    (*F)(1,3) = zDestination - zState;
    (*F)(2,4) = zDestination - zState;

    s->cov() = c.similarity(*F);

    if (fabs(s->x()) < 700);
    else {
      cout << "StateP::predict: produced garbage (lin), vec="
	   << s->vec() << endl;
      //double cc = cin.get();
      return -10;
    } 

  }
  else {
#ifdef TRANSCD
    double fstep = fabs(zDestination - zState);
    if (fstep<=0) fstep = 1;
    double pIn[5];
    pIn[0] = x();
    pIn[1] = y();
    pIn[2] = tx();
    pIn[3] = ty();
    pIn[4] = qPExp;

    double pOut[5];
    int ierror;

    double cOut[25];

    if ( abs(pIn[0]) < 700 &&   // check validity of rkmain input
	 abs(pIn[1]) < 700 &&
	 abs(pIn[2]) < 10   &&
	 abs(pIn[3]) < 10   &&
	 abs(pIn[4]) < 100 );
    else {
      cout << "StateP::predict: invalid pIn="
	   << pIn[0] << " "
	   << pIn[1] << " "
	   << pIn[2] << " "
	   << pIn[3] << " "
	   << pIn[4] 
	   << " transporting from " << zState
	   << " to " << zDestination << endl;
      return -10;
    }

    double cIn[25];
    double rkd[25];
    for (int i=0; i<=4; i++) {
      for (int j=0; j<=4; j++) {
    	cIn[5*i+j] = c(i+1,j+1);
     }
    }

    //rkmain_(icovar,zout,fstep,pin,prec,pOut,rkd,ierror);
    //rktrans_(zState,pin,zDestination,pOut,rkd,ierror);
    double zout = zDestination;
    rktranscd_(zState,pIn,cIn,zout,pOut,cOut,rkd,ierror);

    if (ierror == 0) {
      for (int i=1; i<=5; i++) {
	for (int j=1; j<=5; j++) {
	  (*F)(i,j) = rkd[5*(j-1)+i-1];
	}
      }

      s->setZ() = zDestination;
      s->setX() = pOut[0] + (*F)(1,5) * (qP() - qPExp);
      s->setY() = pOut[1] + (*F)(2,5) * (qP() - qPExp);
      s->setTx() = pOut[2] + (*F)(3,5) * (qP() - qPExp);
      s->setTy() = pOut[3] + (*F)(4,5) * (qP() - qPExp);
      s->setQp() = qP();  // only valid in case momentum is unchanged
      s->setQpEst() = qPEst();

      for (int i=1; i<=5; i++) {
	for (int j=1; j<=5; j++) {
	  s->cov()(i,j) = cOut[5*(j-1)+i-1];
	}
      }

      if (abs(s->x()) < 700);
      else {
	cout << "StateP::predict: produced garbage (mag), vec="
	     << s->vec() << endl;
	cout << " The input was " 
	     << pIn[0] << " "
	     << pIn[1] << " "
	     << pIn[2] << " "
	     << pIn[3] << " "
	     << pIn[4] 
	     << " transporting from " << zState
	     << " to " << zDestination << endl;
	cout << " The output was " 
	     << pOut[0] << " "
	     << pOut[1] << " "
	     << pOut[2] << " "
	     << pOut[3] << endl;
	//double cc = cin.get();
	return -10;
      } 
    }
    else {
      cout << "state::predict: rkmain returns ierror=" << ierror << endl;
      return -1;
    }


#else
    double zout = zDestination;
    double fstep = fabs(zDestination - zState);
    if (fstep<=0) fstep = 1;
    double pin[6];
    pin[0] = zState;
    pin[1] = x();
    pin[2] = y();
    pin[3] = tx();
    pin[4] = ty();
    pin[5] = qPExp;
    double prec[4];
    double dzTrunc = fabs(zDestination - zState);
    if (dzTrunc<1) dzTrunc = 1;

    prec[0] = 0.001 / dzTrunc;
    prec[1] = 0.001 / dzTrunc;
    prec[2] = 0.0001 / dzTrunc;
    prec[3] = 0.0001 / dzTrunc;

    int icovar = 1;
    //int icovar = 2;
    double pout[6];
    double rkd[25];
    int ierror;
    rkmain_(icovar,zout,fstep,pin,prec,pout,rkd,ierror);

    if (ierror == 0) {
      if (pout[0] != zDestination) 
	cout << "stateP::predict: zDest changed from "
	     << zDestination << " to " << pout[0] << endl;
      for (int i=1; i<=5; i++) {
	for (int j=1; j<=5; j++) {
	  (*F)(i,j) = rkd[5*(j-1)+i-1];
	}
      }
      s->setZ() = zDestination;
      s->setX() = pout[1] + (*F)(1,5) * (qP() - qPExp);
      s->setY() = pout[2] + (*F)(2,5) * (qP() - qPExp);
      s->setTx() = pout[3] + (*F)(3,5) * (qP() - qPExp);
      s->setTy() = pout[4] + (*F)(4,5) * (qP() - qPExp);
      s->setQp() = qP();  // only valid in case momentum is unchanged
      s->setQpEst() = qPEst();
    }
    else {
      cout << "state::predict: rkmain returns ierror=" << ierror << endl;
      return -1;
    }
    s->cov() = c.similarity(*F);
#endif



  }


#ifdef FULL
  cout << "rangerStateP::predict: after calculation, *F=" << *F << endl;

  cout << "rangerStateP::predict: the transported vector is"
       << s->vec() << endl;
#endif

  //s->cov() = c.similarity(*F);

#ifdef FULL
  cout << "rangerStateP::predict: the transported cov is"
       << s->cov() << endl;
#endif

  if (1) {                // everything OK
    return 0;
  }
  else {
    cout << " This should not happen!!!" << endl;
    if (newState) delete s;
    if (newTransport) delete F;
    return -1000;
  }
}




int rangerStateP::predictOld(const double& zDestination,
			 const double& qPExp,rangerState*& s,
			 HepMatrix*& F) {

  int newState     = (s == 0);  // denote if memory allocations exist
  int newTransport = (F == 0);   //   .    .    .        .         .

#ifdef FULL
  cout << "rangerStateP::predictOld: zDestination=" << zDestination << endl;

  cout << "rangerStateP::predictOld: targetState=" << s
       << " newState= " <<  newState << endl;
#endif

  if (newState) {
    s = new rangerStateP();
    if (!s) {
      cout << "rangerState::predictOld: allocation error" << endl; 
      exit(1);
    }
    F = new HepMatrix(5,5);
    if (!F) {
      cout << "rangerState::predictOld: allocation error" << endl; 
      exit(1);
    }
  }

#ifdef FULL
  cout << "rangerStateP::predictOld: after allocation: s->vec()=" << s->vec() 
       << endl << "       *F=" << *F << endl;
#endif

  double deltaZ = fabs(zDestination - zState);

  if ( (zDestination>zfpar_.zmaxMag && zState>zfpar_.zmaxMag)
       || (zDestination<zfpar_.zminMag && zState<zfpar_.zminMag)
       || deltaZ < 0.001) {
    s->setZ() = zDestination;      // linear transport
    s->setX() = x() + (zDestination - zState) * tx();
    s->setY() = y() + (zDestination - zState) * ty();
    s->setTx() = tx();
    s->setTy() = ty();
    s->setQp() = qP();
    s->setQpEst() = qPEst();

    for (int i=1; i<=5; i++) {
      for (int j=1; j<=5; j++) {
	(*F)(i,j) = 0.;
      }
    }
    (*F)(1,1) = 1;
    (*F)(2,2) = 1;
    (*F)(3,3) = 1;
    (*F)(4,4) = 1;
    (*F)(5,5) = 1;
    (*F)(1,3) = zDestination - zState;
    (*F)(2,4) = zDestination - zState;
  }
  else {
    double zout = zDestination;
    double fstep = fabs(zDestination - zState);
    if (fstep<=0) fstep = 1;
    double pin[6];
    pin[0] = zState;
    pin[1] = x();
    pin[2] = y();
    pin[3] = tx();
    pin[4] = ty();
    pin[5] = qPExp;
    double prec[4];
    double dzTrunc = fabs(zDestination - zState);
    if (dzTrunc<1) dzTrunc = 1;

    prec[0] = 0.001 / dzTrunc;
    prec[1] = 0.001 / dzTrunc;
    prec[2] = 0.0001 / dzTrunc;
    prec[3] = 0.0001 / dzTrunc;

    int icovar = 1;
    //int icovar = 2;
    double pOut[6];
    double rkd[25];
    int ierror;

    if ( fabs(pin[1]) < 700 &&   // check validity of rkmain input
	 fabs(pin[2]) < 700 &&
	 fabs(pin[3]) < 10   &&
	 fabs(pin[4]) < 10   &&
	 fabs(pin[5]) < 100 );
    else {
      cout << "StateP::predictOld: invalid pin="
	   << pin[1] << " "
	   << pin[2] << " "
	   << pin[3] << " "
	   << pin[4] << " "
	   << pin[5] << endl;
      return -2;
    }


    rkmain_(icovar,zout,fstep,pin,prec,pOut,rkd,ierror);

    if (ierror == 0) {
      if (pOut[0] != zDestination) 
	cout << "state::predictOld: zDest changed from "
	     << zDestination << " to " << pOut[0] << endl;
      for (int i=1; i<=5; i++) {
	for (int j=1; j<=5; j++) {
	  (*F)(i,j) = rkd[5*(j-1)+i-1];
	}
      }
      s->setZ() = zDestination;
      s->setX() = pOut[1] + (*F)(1,5) * (qP() - qPExp);
      s->setY() = pOut[2] + (*F)(2,5) * (qP() - qPExp);
      s->setTx() = pOut[3] + (*F)(3,5) * (qP() - qPExp);
      s->setTy() = pOut[4] + (*F)(4,5) * (qP() - qPExp);
      s->setQp() = qP();  // only valid in case momentum is unchanged
      s->setQpEst() = qPEst();
    }
    else {
      cout << "state::predictOld: rkmain returns ierror=" << ierror << endl;
      return -1;
    }
  }


#ifdef FULL
  cout << "rangerStateP::predictOld: after calculation, *F=" << *F << endl;

  cout << "rangerStateP::predictOld: the transported vector is"
       << s->vec() << endl;
#endif

  s->cov() = c.similarity(*F);

#ifdef FULL
  cout << "rangerStateP::predictOld: the transported cov is"
       << s->cov() << endl;
#endif

  if (1) {                // everything OK
    return 0;
  }
  else {
    cout << " This should not happen!!!" << endl;
    if (newState) delete s;
    if (newTransport) delete F;
    return -1000;
  }
}


void rangerStateP::applyScat(const rangerScatThin& scat) {

  cov()(3,3) += scat.cTxTx;
  cov()(4,4) += scat.cTyTy;
  cov()(3,4) += scat.cTxTy;   // exploit symmetric matrix

  float xField[3];
  xField[0] = x();
  xField[1] = y();
  xField[2] = zState;
  float hField[3];
  utfeld_(xField,hField);
  
  if (zState > zfpar_.zminMag && zState < zfpar_.zmaxMag) {
    if (fabs(hField[1]) > 0.1) {
      cov()(5,5) += zfpar_.fP * fabs(hField[1]) * scat.dScat;
    }
  }
}

void rangerStateP::applyScat(const rangerScatThick& scat) {

  cov()(3,3) += scat.cTxTx;
  cov()(4,4) += scat.cTyTy;
  cov()(3,4) += scat.cTxTy;   // exploit symmetric matrix

  cov()(1,1) += scat.cXX;     // correction for 
  cov()(1,2) += scat.cXY;     // thick scatterer
  cov()(2,2) += scat.cYY;
  cov()(1,3) += scat.cXTx;
  cov()(1,4) += scat.cXTy;
  cov()(2,3) += scat.cYTx;
  cov()(2,4) += scat.cYTy;

  float xField[3];
  xField[0] = x();
  xField[1] = y();
  xField[2] = zState;
  float hField[3];
  utfeld_(xField,hField);
  
  if (zState > zfpar_.zminMag && zState < zfpar_.zmaxMag) {
    if (fabs(hField[1]) > 0.1) {
      cov()(5,5) += zfpar_.fP * fabs(hField[1]) * scat.dScat;
    }
  }
}


void rangerStateP::applyRadEl(const rangerRadEl& rad) {
  if (zState > zfpar_.zminEc && zState < zfpar_.zmaxEc) {
    p(5) *= rad.qPFactor;
    cov()(5,5) += rad.cQpQp;
  }
}


void rangerStateP::applyDeDx(const float rl) {
  if (zState < zfpar_.zmaxMag && qP() != 0) {
    //float eLoss = 0.075 * rl;
    float eLoss = 0.048 * rl;     // tuned for radThickEff
    if (qP() != 0) {
      setQp() = (qP() > 0)? (1./(1./qP() + eLoss)) : (1./(1./qP() - eLoss));
    }
  }
}

void rangerStateP::applyRadPhoton(const float e,const float dE) {
  if (qP() != 0) {
    double qPOld = qP();
    setQp() = (qP() > 0)? (1./(1./qP() + e)) : (1./(1./qP() - e));
    double efac = qP() / qPOld;
    double dCov = (efac * efac * efac * efac - 1) * cov()(5,5)
      + dE * dE * qP() * qP() * qP() * qP();
    if (dCov >0) cov()(5,5) += dCov;
  }
}


void rangerStateP::storeRseg(ArtePointer<RSEG> rseg,int firstLast){
 switch (firstLast) {
 case 1:
   rseg->zf = zState;
   rseg->xf = x();
   rseg->yf = y();
   rseg->txf = tx();
   rseg->tyf = ty();
   rseg->pf = qP();
   rseg->cvf[ 0] = cov()(1,1);
   rseg->cvf[ 1] = cov()(1,2);
   rseg->cvf[ 2] = cov()(2,2);
   rseg->cvf[ 3] = cov()(3,1);
   rseg->cvf[ 4] = cov()(3,2);
   rseg->cvf[ 5] = cov()(3,3);
   rseg->cvf[ 6] = cov()(4,1);
   rseg->cvf[ 7] = cov()(4,2);
   rseg->cvf[ 8] = cov()(4,3);
   rseg->cvf[ 9] = cov()(4,4);
   rseg->cvf[10] = cov()(5,1);
   rseg->cvf[11] = cov()(5,2);
   rseg->cvf[12] = cov()(5,3);
   rseg->cvf[13] = cov()(5,4);
   rseg->cvf[14] = cov()(5,5);
   rseg->time = 0;
   rseg->dtim = 0;
   break;
 case 2:
   rseg->ze = zState;
   rseg->xe = x();
   rseg->ye = y();
   rseg->txe = tx();
   rseg->tye = ty();
   rseg->pe = qP();
   rseg->cve[ 0] = cov()(1,1);
   rseg->cve[ 1] = cov()(1,2);
   rseg->cve[ 2] = cov()(2,2);
   rseg->cve[ 3] = cov()(3,1);
   rseg->cve[ 4] = cov()(3,2);
   rseg->cve[ 5] = cov()(3,3);
   rseg->cve[ 6] = cov()(4,1);
   rseg->cve[ 7] = cov()(4,2);
   rseg->cve[ 8] = cov()(4,3);
   rseg->cve[ 9] = cov()(4,4);
   rseg->cve[10] = cov()(5,1);
   rseg->cve[11] = cov()(5,2);
   rseg->cve[12] = cov()(5,3);
   rseg->cve[13] = cov()(5,4);
   rseg->cve[14] = cov()(5,5);
   break;
 default:
   cout << "rangerStateP::storeRseg: invalid firstLast " << firstLast << endl;
   exit(1);
   break;
 }
}


void rangerStateP::storeRtra(ArtePointer<RTRA> rtra,int firstLast){
 switch (firstLast) {
 case 1:
   rtra->zf = zState;
   rtra->xf = x();
   rtra->yf = y();
   rtra->txf = tx();
   rtra->tyf = ty();
   rtra->pf = qP();
   rtra->cvf[ 0] = cov()(1,1);
   rtra->cvf[ 1] = cov()(1,2);
   rtra->cvf[ 2] = cov()(2,2);
   rtra->cvf[ 3] = cov()(3,1);
   rtra->cvf[ 4] = cov()(3,2);
   rtra->cvf[ 5] = cov()(3,3);
   rtra->cvf[ 6] = cov()(4,1);
   rtra->cvf[ 7] = cov()(4,2);
   rtra->cvf[ 8] = cov()(4,3);
   rtra->cvf[ 9] = cov()(4,4);
   rtra->cvf[10] = cov()(5,1);
   rtra->cvf[11] = cov()(5,2);
   rtra->cvf[12] = cov()(5,3);
   rtra->cvf[13] = cov()(5,4);
   rtra->cvf[14] = cov()(5,5);
   break;
 case 2:
   break;
 default:
   break;
 }
}


void rangerStateP::storeRhit(ArtePointer<RHIT> rhit) {
  rhit->x = x();
  rhit->y = y();
  rhit->tx = tx();
  rhit->ty = ty();
  rhit->p = qP();
}

ArtePointer<RPNT> rangerStateP::storeRpnt(ArtePointer<RPNT> rpnt){
  ArtePointer<RPNT> rp;
  if (rpnt == 0) {
    // create new entry in RPNT. 
    mmbook_(ArtePointer<RPNT>::getLink().getRef(),1);
    rp = ArtePointer<RPNT>( ArtePointer<RPNT>::nrow() );
  }
  else {
    rp = rpnt;
  }
  //cout << "rangerStateP::storeRpnt: store in rpnt=" << rp << endl;
  if (rp != 0) {
    rp->zf = zState;
    rp->xf = x();
    rp->yf = y();
    rp->txf = tx();
    rp->tyf = ty();
    rp->pf = qP();
    rp->cvf[ 0] = cov()(1,1);
    rp->cvf[ 1] = cov()(1,2);
    rp->cvf[ 2] = cov()(2,2);
    rp->cvf[ 3] = cov()(3,1);
    rp->cvf[ 4] = cov()(3,2);
    rp->cvf[ 5] = cov()(3,3);
    rp->cvf[ 6] = cov()(4,1);
    rp->cvf[ 7] = cov()(4,2);
    rp->cvf[ 8] = cov()(4,3);
    rp->cvf[ 9] = cov()(4,4);
    rp->cvf[10] = cov()(5,1);
    rp->cvf[11] = cov()(5,2);
    rp->cvf[12] = cov()(5,3);
    rp->cvf[13] = cov()(5,4);
    rp->cvf[14] = cov()(5,5);
    return rp;
  }
  else {
    return 0;
  }
}


void rangerStateP::copyVec( const rangerState& s) {
  zState = s.z();
  p = s.vec();
  qPEstimate = s.qPEst();
}




void rangerStateP::linExtrapVec( const rangerState& s, float z) {

  zState = z;

  setX() = s.x() + (z - s.z()) * s.tx();
  setY() = s.y() + (z - s.z()) * s.ty();
  setTx() =  s.tx();
  setTy() =  s.ty();
  setQp() =  s.qP();
  setQpEst() =  s.qPEst();
}

void rangerStateP::dump() {
  cout << "     zState=" << zState << "  vec=" << p << endl;
  cout << "                              cov=" << c << endl;
}
