// 
//  Author:  Rainer Mankel                     Created:  29-Sep-98
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// 
#include <iostream.h>
#include <algo.h>
#include "arte/arte.hh"
#include "arte/RSEG_cmp.hh"
#include "rang/Node.hh"
#include "rang/Hit.hh"
#include "rang/HitWire.hh"
#include "rang/Geom.hh"
#include "rang/Volume.hh"
#include "rang/Track.hh"
#include "rang/hfill.hh"
#include "rang/zfpar.hh"
#include "rang/RMap.hh"
#include "rang/HitMap.hh"
#include "rang/FLAGHist.hh"

#include <list.h>
#include <vector.h>


#undef SAVEWALLS

//-- Default constructor
rangerTrack::rangerTrack() {
  chiSquare = 0;
  qpEstimate = 0;
  electron = 0;
  nIteration = 0;
  smoothed = 0;
  tFitMode = ModeP;
  tPartType = MUON;
  status = 0;
};

//-- constructor from RSEG
rangerTrack::rangerTrack(ArtePointer<RSEG> rseg, rangerFitMode fm,
			 rangerParticleType pType,
			 rangerRMap* rmap) {
  chiSquare = 0;
  qpEstimate = rseg->pf;
  electron = 0;        // set according to electron hypothesis
  nIteration = 0;
  smoothed = 0;
  tFitMode = fm;
  tPartType = pType;
  status = 0;

  // the RHITs are assumed to be retrieved in the order of increasing z.
  // the nodes are stored such that the first node in the list 
  // points to highest z
  
  for ( ArteRelationIterator<RHIT> rhitIt = rseg->rhit.begin();
	rhitIt != rseg->rhit.end(); ++rhitIt ) {
    ArtePointer<RHIT> rhit(*rhitIt);
#ifdef FULL
    cout << "Current rhit=" << rhit << 
      " hitb=" << rhit->hitb << " z=" << rhit->hitb->z << endl;
#endif

    rangerNode* w = rangerNode::makeNode( rhit, fitMode() );
    if (w) {
      prepend(*w);
      if (rmap) {
	rangerHitMap hm( w,rhit );
	rmap->addHitMap( hm );
      }
    }
  }
};

//-- constructor from RSEG vector
rangerTrack::rangerTrack(const vector<ArtePointer<RSEG> >& rsegVec, 
			 rangerFitMode fm,
			 rangerParticleType pType,
			 vector<rangerRMap>* rmapVec) {
  chiSquare = 0;
  qpEstimate = 0;
  electron = 0;        // set according to electron hypothesis
  nIteration = 0;
  smoothed = 0;
  tFitMode = fm;
  tPartType = pType;
  status = 0;

  // the RHITs are assumed to be retrieved in the order of increasing z.
  // the nodes are stored such that the first node in the list 
  // points to highest z

  for (vector<ArtePointer<RSEG> >::const_iterator sIt=rsegVec.begin();
       sIt != rsegVec.end(); sIt++) {
    ArtePointer<RSEG> rseg = *sIt;
    if (rseg->pf != 0) qpEstimate = rseg->pf;
    rangerRMap rMap(rseg);
  
    for ( ArteRelationIterator<RHIT> rhitIt = rseg->rhit.begin();
	  rhitIt != rseg->rhit.end(); ++rhitIt ) {
      ArtePointer<RHIT> rhit(rhitIt);
#ifdef FULL
      cout << "Track(Vec...): current rhit=" << rhit << 
	" hitb=" << rhit->hitb << " z=" << rhit->hitb->z << endl;
#endif

      rangerNode* w = rangerNode::makeNode( rhit, fitMode() );
      if (w) {
	prepend(*w);
	rangerHitMap hm( w,rhit );
	rMap.addHitMap( hm );
      }
    }
    if (rmapVec) rmapVec->push_back(rMap);

  }
};


//-- constructor from RSEG vector + RCCL
rangerTrack::rangerTrack(const vector<ArtePointer<RSEG> >& rsegVec, 
			 ArtePointer<RCCL> rccl,int charge,
			 rangerFitMode fm,
			 rangerParticleType pType,
			 vector<rangerRMap>* rmapVec) {
  chiSquare = 0;
  qpEstimate = 0;
  electron = 0;        // set according to electron hypothesis
  nIteration = 0;
  smoothed = 0;
  tFitMode = fm;
  tPartType = pType;
  status = 0;

  // the RHITs are assumed to be retrieved in the order of increasing z.
  // the nodes are stored such that the first node in the list 
  // points to highest z

  for (vector<ArtePointer<RSEG> >::const_iterator sIt=rsegVec.begin();
       sIt != rsegVec.end(); sIt++) {
    ArtePointer<RSEG> rseg = *sIt;
    if (rseg->pf != 0) qpEstimate = rseg->pf;
    rangerRMap rMap(rseg);
  
    for ( ArteRelationIterator<RHIT> rhitIt = rseg->rhit.begin();
	  rhitIt != rseg->rhit.end(); ++rhitIt ) {
      ArtePointer<RHIT> rhit(*rhitIt);
#ifdef FULL
      cout << "Track(Vec...): current rhit=" << rhit << 
	" hitb=" << rhit->hitb << " z=" << rhit->hitb->z << endl;
#endif

      rangerNode* w = rangerNode::makeNode( rhit, fitMode() );
      if (w) {
	prepend(*w);
	rangerHitMap hm( w,rhit );
	rMap.addHitMap( hm );
      }
    }
    if (rmapVec) rmapVec->push_back(rMap);

  }
  rangerNode* w = rangerNode::makeNode( rccl,charge,fitMode());
  if (w) prepend(*w);

};


//-- constructor from RTRA
rangerTrack::rangerTrack(ArtePointer<RTRA> rtra, rangerFitMode fm,
			 rangerParticleType pType,
			 rangerRMap* rmap) {
  chiSquare = 0;
  qpEstimate = rtra->pf;
  electron = 0;        // set according to electron hypothesis
  nIteration = 0;
  smoothed = 0;
  tFitMode = fm;
  tPartType = pType;
  status = 0;

  // the RHITs are assumed to be retrieved in the order of increasing z.
  // the nodes are stored such that the first node in the list 
  // points to highest z
  
  //ArtePointer<RTRX> rtrx = rtra;

  for ( ArteRelationIterator<HITB> hitbIt = rtra->hitb.begin();
	hitbIt != rtra->hitb.end(); ++hitbIt ) {
    ArtePointer<HITB> hitb(*hitbIt);
#ifdef FULL
    cout << //"Current rhit=" << rhit << 
      " hitb=" << *hitbIt << " z=" << (*hitbIt)->z << endl;
#endif

    for ( ArteRelationIterator<RHIT> rhitIt = hitb->rhit.begin();
	rhitIt != hitb->rhit.end(); ++rhitIt ) {
      ArtePointer<RHIT> rhit(*rhitIt);
      if (rhit->rtra == rtra) {
	rangerNode* w = rangerNode::makeNode( rhit, fitMode() );
	if (w) {
	  prepend(*w);
	  if (rmap) {
	    rangerHitMap hm( w,rhit );
	    rmap->addHitMap( hm );
	  }
	}
      }
    }
  }
}


// Destructor
rangerTrack::~rangerTrack() {
  for (rangerNodeIterator nodeIt = begin();
       nodeIt != end();
       nodeIt++) {
    delete *nodeIt;
  }
}

ostream& operator<<(ostream &os,const rangerTrack &t) {
  os << " #nodes=" << (t.nodes).size() << " chiSquare=" << t.chiSquare
     << " qpEst=" << t.qpEstimate
     << " FMode=" << int(t.tFitMode)
     << " ParType=" << int(t.tPartType) << endl;
  for (list<rangerNode*>::const_iterator n = (t.nodes).begin(); 
       n != (t.nodes).end(); n++) {
    os << *n << '\t' << *(*n) << endl;
  }
  return os;
}
  

void rangerTrack::covInit() {
  (*begin())->covInit();
}



int rangerTrack::fit(const int firstPass,const int thirdPass,
		     const int insertMaterial,const int updateInternal) {

  // PASS 1
  if (firstPass) {
    status = filter();
    if (status < 0) {
      return -10;
    }

    status = smooth();
    
    if (status < 0) {
      return -11;
    }
    if (updateInternal) updInternal();
  }

  int nInsert = 0;
  if (insertMaterial) {
    nInsert=insertWalls();
  }

  // PASS 2
  if (nInsert || (! firstPass)) {

    status = filter();
    if (status < 0) {
      return -11;
    }
    status = smooth();
    if (status < 0) {
      return -11;
    }
    if (updateInternal) updInternal();

    // PASS 3
    if (thirdPass) {
      if (fabs((*rbegin())->getState().qP()) > 0.19) {
	status = filter();
	if (status < 0) {
	  return -12;
	}
	status = smooth();
	if (updateInternal) updInternal();
	if (status < 0) {
	  return -11;
	}
      }
    }
  }
  return 0;
}


int rangerTrack::fitFast() {

  status = filterDownStr();
  if (status < 0) {
    return -10;
  }
  status = filter();
  if (status < 0) {
    return -10;
  }
  return 0;
}




int rangerTrack::filter() { 

#ifdef FULL
  int counter=0;
#endif

  if (! empty()) {
    int stat = (*begin())->filter();

#ifdef FULL
    cout << "first filter status stat=" << stat << endl;

    cout << "rangerTrack::filter: after filter step " << ++counter << endl;
    //char cc;
    //cc = cin.get();
#endif

    if (stat >= 0) {
      int last_stat = 0;
      rangerNodeIterator baseNode = begin();
      rangerNodeIterator nextNode = begin();
      nextNode++;
      for (rangerNodeIterator n = nextNode; 
	   n != end() && (last_stat >=0); 
	   n++) {
	last_stat = (*n)->filter(*baseNode,tPartType);
	baseNode = n;
      }
      return last_stat;
    } 
    else {
      return stat;
    }
  }
  cout << "ranger::Track.filter(): no first node" << endl;
  return -10;
}



      
int rangerTrack::smooth() { 

  if ( empty() ) {
    cout << "rangerTrack::smooth(): empty track" << endl;
    return -10;
  }

  int last_stat = 0;
  for (rangerNodeReverseIterator n = rbegin(); 
       (n != rend()) && (last_stat >= 0); n++) {
    rangerNodeReverseIterator nk = n;
    nk++;
    if (nk != rend()) last_stat = (*nk)->smooth(*n);
  }
  return last_stat;
}

int rangerTrack::filterDownStr() { 

#ifdef FULL
  int counter=0;
#endif

  if (! empty()) {
    int stat = (*rbegin())->filter();

#ifdef FULL
    cout << "first filterDownStr status stat=" << stat << endl;

    cout << "rangerTrack::filterDownStr: after filter step " 
	 << ++counter << endl;
#endif

    if (stat >= 0) {
      int last_stat = 0;
      rangerNodeReverseIterator baseNode = rbegin();
      rangerNodeReverseIterator nextNode = rbegin();
      nextNode++;
      for (rangerNodeReverseIterator n = nextNode; 
	   n != rend() && (last_stat >=0); 
	   n++) {
	last_stat = (*n)->filter(*baseNode,tPartType);
	baseNode = n;
#ifdef FULL
	cout << "filter status last_stat=" << last_stat << endl;
	cout << "rangerTrack::filterDownStr: after filter step " 
	     << ++counter << endl;
	//cc = cin.get();
#endif
      }
      if (last_stat != 0) 
	cout << "rangerTrack::filterDownStr: filtering returns last_stat=" 
	     << last_stat << endl;
      return last_stat;
    } 
    else {
      cout << "rangerTrack::filterDownStr: first filtering returns stat=" 
	   << stat << endl;
      return stat;
    }
  }
  cout << "ranger::Track.filterDownStr(): no first node" << endl;
  return -10;
}
      


int rangerTrack::insertWalls(){
  int nInserts = 0;
  if ( empty() ) return 0;
  // test if first node corresponds to highest z
  if ((*begin())->zNode() < (*last())->zNode()) {
    cout << "rangerTrack::insertWalls(): firstNode->zNode()="
	 << (*begin())->zNode() << " last->zNode()="
	 << (*last())->zNode() << endl;
    return 0;
  }

  for (rangerNodeIterator nodeIt=begin(); nodeIt != end(); ) {
    rangerNodeIterator followingNode =  next(nodeIt);
    rangerNodeIterator insertBaseNode = nodeIt;
    if (next(nodeIt) != end()) {   // locate in-between nodes
      int beginPlane = (*nodeIt)->plane();
      int endPlane = (*next(nodeIt))->plane();
      if ( beginPlane > 0 && endPlane > 0 && beginPlane != endPlane ) {
	ArtePointer<GEDE> beginGede(beginPlane);
	rangerVolume* beginVolume = rangerGeom::vol( beginGede );
	ArtePointer<GEDE> endGede(endPlane);
	rangerVolume* endVolume = rangerGeom::vol( endGede );
#ifdef FULL
	cout << " search between " << beginPlane 
	     << " z=" << beginVolume->zMean() << " zRef=" << beginVolume->zRef()
	     << " and " << endPlane 
	     << " z=" << endVolume->zMean() << " zref=" << endVolume->zRef()
	     << endl;
#endif
	for ( rangerVolume* volIt=beginVolume->previous() ;
	      volIt && (volIt != endVolume) 
		&& (volIt->zRef() > endVolume->zRef());
	      volIt = volIt->previous()) {
	  if (volIt->zRef() < beginVolume->zRef()) {
#ifdef FULL
	    cout << " located volume " << volIt->plNumber() 
		 << " zRef=" << volIt->zRef();
	    if (volIt->previous() == endVolume) cout << "[last one]";
	    else if (volIt->previous()) cout << "[next " 
					     <<  (volIt->previous())->plNumber()
					     << " "
					     << (volIt->previous())->zRef()
					     << "]";
	    cout << endl;
#endif
	    if (float sLeng = volIt->sLengNode(*nodeIt)) {
#ifdef INSERTPRINT
	      cout << " -------- passed volume " << volIt->plNumber() 
		   << " at z=" << volIt->zMean() << 
		" between " << beginVolume->plNumber()
		   << "/" << beginVolume->zMean()
		   << " and " << endVolume->plNumber()
		   << "/" << endVolume->zMean() << endl;
#endif
	      rangerNode* newNode = 0;
#ifdef INSERTPRINT
	      cout << "Track::insert: getGede=" << volIt->getGede()
		   << " getGwal=" << volIt->getGwal() 
		   << " ms=" << volIt->ms() << " z=" << volIt->zMean()
		   << " shape " << volIt->getGeWa()->shp() << endl;
#endif
	      if (volIt->getGede() != 0) {
		newNode = rangerNode::makeNode(volIt->getGede(),*nodeIt,sLeng);
	      }
	      else {
		if (volIt->getGwal() != 0) {
		  newNode = rangerNode::makeNode(volIt->getGwal(),*nodeIt,sLeng);
		}
	      }
	      // insert node into track
	      if (newNode) {
		insertBaseNode = insertAfter(insertBaseNode,newNode);
		nInserts++;
	      }
	    }
	  }
#ifdef INSERTPRINT
	  cout << " ........................";
	  if (volIt->previous()) 
	    cout << "[prev OK]" << "[ne endvol " << (volIt->previous() != endVolume)
		 << "][zref " << (volIt->previous())->zRef() << ">"
		 << endVolume->zRef() << "? "
		 << ((volIt->previous())->zRef() > endVolume->zRef())
		 << endl;
	  else cout << "[prev bad]" << endl;
#endif
	}
      }
    }
    nodeIt = followingNode;
  }
  return nInserts;
}



void rangerTrack::updInternal(){
  for (rangerNodeIterator nodeIt = begin();
       nodeIt != end();
       nodeIt++) {
    (*nodeIt)->updInternal();
  }
}



void rangerTrack::append(rangerNode& node) {
  nodes.push_back( &node );
}

void rangerTrack::prepend(rangerNode& node) {
  nodes.push_front( &node );
}


list<rangerNode*>::iterator rangerTrack::insertAfter
                                 (rangerNodeIterator baseNode, 
                                            rangerNode* newNode) { 
  if (baseNode != end()) {
    rangerNodeIterator insertPosition = baseNode;
    insertPosition++;
    rangerNodeIterator tmp = nodes.insert( insertPosition,newNode);
    return tmp;
  }
  else {
    cout << "rangerTrack::insertAfter: invalid base node " << endl;
    return baseNode;
  }
} 


void rangerTrack::trackList() {
  cout << "tFitMode=" << int(tFitMode) << endl;
  for (rangerNodeIterator n = begin(); n != end(); n++) {
    (*n)->list();
  }
}


void rangerTrack::dump() {
  cout << "D u m p           tFitMode=" << int(tFitMode) << endl;
  for (rangerNodeIterator n = begin(); n != end(); n++) {
    (*n)->dump();
  }
}


void rangerTrack::storeback(rangerRMap* rmap){

  if (! rmap) return;
  if (empty()) return;

  // Pass 0: prepare RPNTs for particle id
  rmap->prepareIdPnt();

  // for RTRA: prepare RPNT for high z point
  if ((rmap->getRtra()) != 0) {
    rmap->addPnt(*begin());
  }

  // Pass 1: store the RHITs
  for (rangerNodeIterator n = begin(); n != end(); n++) {
    if ((*n)->hasHitb()) {
      ArtePointer<RHIT> rhit = rmap->getRhit(*n);
      if (rhit != 0) (*n)->storeRhit(rhit);
    }
#ifdef SAVEWALLS
    else {              // store wall RPNTs (for diagnosis only)
      if (rmap->getRpnt(*n) == 0) {
	ArtePointer<RPNT> rp = (*n)->storeRpnt(0);
	if (rp != 0) {  // new RPNT, set relation
	  if ((rmap->getRseg()) != 0) {
	    createRelation( rmap->getRseg(), rp, &RSEG::rpnt );
// 	    mmbrel_(ArtePointer<RSEG_RPNT>::getLink().getRef(),
// 		    (rmap->getRseg()).index(),rp.index());
	  }
	  else if ((rmap->getRtra()) != 0) {
	    createRelation( rmap->getRtra(), rp, &RTRA::rpnt );
// 	    mmbrel_(ArtePointer<RTRX_RPNT>::getLink().getRef(),
// 		    (rmap->getRtra()).index(),rp.index());
	  }
	}
      }
    }
#endif
  }

  // Pass 2A: store the RSEG
  if ((rmap->getRseg()) != 0) {
    rangerNode* nLow = rmap->getNodeZLow();
    if (nLow) nLow->storeRseg(rmap->getRseg(),1);
    rangerNode* nHigh = rmap->getNodeZHigh();
    if (nHigh) nHigh->storeRseg(rmap->getRseg(),2);

    (rmap->getRseg())->chi = rmap->chiSq();

    (rmap->getRseg())->fit = (status<0)? -10: 0;

    if ((rmap->getRseg())->cmp == 0) (rmap->getRseg())->cmp = genCmp();
  }
  // Pass 2B: store the RTRA
  else if ((rmap->getRtra()) != 0) {
    rangerNode* nLow = rmap->getNodeZLow();
    if (nLow) nLow->storeRtra(rmap->getRtra(),1);

    (rmap->getRtra())->fit = (status<0)? -10: 0;
    (rmap->getRtra())->chi = rmap->chiSq();
  }

  // pass 3: store the RPNT
  for ( list< rangerHitMap >::const_iterator 
	  HMIt=(rmap->getHMList()).begin();
	HMIt !=(rmap->getHMList()).end(); 
	HMIt++){
    if ((*HMIt).getRpnt() != 0) {   // store into assigned RPNT
      (*HMIt).getNode()->storeRpnt((*HMIt).getRpnt());

      if ((rmap->getRseg()) != 0) {
	int related = 0;
	related = (((*HMIt).getRpnt())->rseg == rmap->getRseg());

	if (! related) {  // set relation
	  createRelation( rmap->getRseg(), (*HMIt).getRpnt(), &RSEG::rpnt );
// 	  mmbrel_(ArtePointer<RSEG_RPNT>::getLink().getRef(),
// 		  (rmap->getRseg()).index(),((*HMIt).getRpnt()).index());
	}
      }
      else if ((rmap->getRtra()) != 0) {
	if( find( (*HMIt).getRpnt()->rtra.begin(),
		  (*HMIt).getRpnt()->rtra.end(),
		  rmap->getRtra() ) != (*HMIt).getRpnt()->rtra.end() ) {
	  createRelation( rmap->getRtra(), (*HMIt).getRpnt(), &RTRA::rpnt );
	}
//	int related = 0;
// 	for (ArteRelationIterator<RPNT_RTRA> rtraIt(((*HMIt).getRpnt())->rtra);
// 	     rtraIt.valid();
// 	     rtraIt.next()) {
// 	  if ( ( (ArtePointer<RTRA>) rtraIt) == rmap->getRtra()) {
// 	    related = 1;
// 	    break;
// 	  }
// 	}
// 	if (! related) {  // set relation
// 	  mmbrel_(ArtePointer<RTRX_RPNT>::getLink().getRef(),
// 		  (rmap->getRtra()).index(),((*HMIt).getRpnt()).index());
// 	}
      }
    }
  }
};


void rangerTrack::storeback(vector<rangerRMap>* rmapVec) {
  if (! rmapVec) return;
  for (vector<rangerRMap>::iterator rIt=rmapVec->begin();
       rIt != rmapVec->end(); rIt++) {
    storeback( &(*rIt) );
  }
}


ArtePointer<RTRA> rangerTrack::storeRtra() const{

  if (empty()) return 0;

  // Pass 1: allocate and store RTRA entry
  mmbook_(ArtePointer<RTRA>::getLink().getRef(),1);
  ArtePointer<RTRA> rt( ArtePointer<RTRA>::nrow() );
  (*rbegin())->storeRtra(rt,1);   // store point of lowest z

  rt->fit = (status<0)? -10: 0;
  rt->chi = totalChiSq();
  rt->hsi = 0;                  // Initialize hit counts
  rt->hin = 0;
  rt->hou = 0;
  rt->hri = 0;
  rt->hca = 0;
  rt->htr = 0;
  rt->hmu = 0;

  rangerRMap rmap(rt);  // create hit map to aid further operations

  // Pass 2: store the RHITs
  for (rangerNodeConstIterator n = begin(); n != end(); n++) {
    if ((*n)->hasHitb()) {
      mmbook_(ArtePointer<RHIT>::getLink().getRef(),1);
      ArtePointer<RHIT> rhit( ArtePointer<RHIT>::nrow() );
      if (rhit != 0) {
	createRelation( rt, (*n)->getHitb(), &RTRA::hitb );
	createRelation( rhit, (*n)->getHitb(), &HITB::rhit );
// 	mmbrel_(ArtePointer<RTRX_HITB>::getLink().getRef(),
// 		rt.index(),((*n)->getHitb()).index());
// 	mmbrel_(ArtePointer<HITB_RHIT>::getLink().getRef(),
// 		((*n)->getHitb()).index(),rhit.index());
	(*n)->storeRhit(rhit);
	rhit->rtra = rt.index();
	rangerHitMap hm((*n),rhit);
	rmap.addHitMap( hm );
      }
      ArtePointer<GEDE> ge(((*n)->getHitb())->gede);  // count hits
      switch (ge->cmp) {
      case 1:
	rt->hsi++;
	break;
      case 2:
	rt->hou++;
	break;
      case 3:
	rt->hin++;
	break;
      case 4:
	rt->hri++;
	break;
      case 5:
	rt->hca++;
	break;
      case 6:
	rt->htr++;
	break;
      case 7:
	rt->hmu++;
	break;
      default:
	cout << " rangerTrack::storeRtra: bad cmp=" << ge->cmp << endl;
      }
    }
#ifdef SAVEWALLS
    else {              // store wall RPNTs (for diagnosis only)
      ArtePointer<RPNT> rp = (*n)->storeRpnt(0);
	if (rp != 0) {  // new RPNT, set relation
	  createRelation( rt, rp, &RTRA::rpnt );
// 	  mmbrel_(ArtePointer<RTRX_RPNT>::getLink().getRef(),
// 		  rt.index(),rp.index());
	}
    }
#endif
  }
  // Pass 3: prepare RPNTs for particle id
  rmap.prepareIdPnt();

  // prepare RPNT for high z point
  rmap.addPnt(*begin());

  // pass 4: store the RPNT
  for ( list< rangerHitMap >::const_iterator 
	  HMIt=(rmap.getHMList()).begin();
	HMIt !=(rmap.getHMList()).end(); 
	HMIt++){
    if ((*HMIt).getRpnt() != 0) {   // store into assigned RPNT
      (*HMIt).getNode()->storeRpnt((*HMIt).getRpnt());

      // set relation
      createRelation( rt, (*HMIt).getRpnt(), &RTRA::rpnt );
//       mmbrel_(ArtePointer<RTRX_RPNT>::getLink().getRef(),
// 		  rt.index(),((*HMIt).getRpnt()).index());
    }
  }
  return rt;
}


ArtePointer<RSEG> rangerTrack::storeRseg() const{

  if (empty()) return 0;

  // Pass 1: allocate and store RSEG entry
  mmbook_(ArtePointer<RSEG>::getLink().getRef(),1);
  ArtePointer<RSEG> rs( ArtePointer<RSEG>::nrow() );
  (*rbegin())->storeRseg(rs,1);   // store point of lowest z
  (*begin())->storeRseg(rs,2);    // store point of high z

//   if (rs->cvf[14]<0.5*zfpar_.cqqrf) {  // set fit flag
//     rs->fit = 2;
//   }
//   else {
//     rs->fit = 1;
//   }
  rs->fit = (status<0)? -10: 0;
  rs->chi = totalChiSq();
  rs->cmp = genCmp();

  rangerRMap rmap(rs);  // create hit map to aid further operations

  // Pass 2: store the RHITs
  for (rangerNodeConstIterator n = begin(); n != end(); n++) {
    if ((*n)->hasHitb()) {
      mmbook_(ArtePointer<RHIT>::getLink().getRef(),1);
      ArtePointer<RHIT> rhit( ArtePointer<RHIT>::nrow() );
      if (rhit != 0) {
	createRelation( rs,   (*n)->getHitb(), &RSEG::hitb );
	createRelation( rhit, (*n)->getHitb(), &HITB::rhit );
	createRelation( rs,   rhit,            &RSEG::rhit );
// 	mmbrel_(ArtePointer<RSEG_HITB>::getLink().getRef(),
// 		rs.index(),((*n)->getHitb()).index());
// 	mmbrel_(ArtePointer<HITB_RHIT>::getLink().getRef(),
// 		((*n)->getHitb()).index(),rhit.index());
// 	mmbrel_(ArtePointer<RSEG_RHIT>::getLink().getRef(),
// 		rs.index(),rhit.index());
	(*n)->storeRhit(rhit);
	rhit->rtra = 0;
	rangerHitMap hm((*n),rhit);
	rmap.addHitMap( hm );
      }
    }
#ifdef SAVEWALLS
    else {              // store wall RPNTs (for diagnosis only)
      ArtePointer<RPNT> rp = (*n)->storeRpnt(0);
	if (rp != 0) {  // new RPNT, set relation
	  createRelation( rs, rp, &RSEG::rpnt );
// 	  mmbrel_(ArtePointer<RSEG_RPNT>::getLink().getRef(),
// 		  rs.index(),rp.index());
	}
    }
#endif
  }

  // Pass 3: prepare RPNTs for particle id
  rmap.prepareIdPnt();

  // pass 4: store the RPNT
  for ( list< rangerHitMap >::const_iterator 
	  HMIt=(rmap.getHMList()).begin();
	HMIt !=(rmap.getHMList()).end(); 
	HMIt++){
    if ((*HMIt).getRpnt() != 0) {   // store into assigned RPNT
      (*HMIt).getNode()->storeRpnt((*HMIt).getRpnt());

      // set relation
      createRelation( rs, (*HMIt).getRpnt(), &RSEG::rpnt );
//       mmbrel_(ArtePointer<RSEG_RPNT>::getLink().getRef(),
// 		  rs.index(),((*HMIt).getRpnt()).index());
    }
  }
  return rs;
}






double rangerTrack::getLowZ() const {
  if (empty()) return 0;
  else {
    return (*rbegin())->zNode();
  }
}

const HepVector* rangerTrack::getLowZVec() const {
  if (empty()) return 0;
  else {
    return &(((*rbegin())->getState()).vec());
  }
}

const HepSymMatrix* rangerTrack::getLowZCov() const {
  if (empty()) return 0;
  else {
    return &(((*rbegin())->getState()).cov());
  }
}

double const rangerTrack::totalChiSq() const {
  
  double sum = 0;
  int nCoord=0;
  for (rangerNodeConstIterator n = begin(); n != end() ; n++) {
    sum += (*n)->getChiSq();
    nCoord += (*n)->dim();  //count dimension of measurements

#ifdef rangerHist
    int id = 7001;
    float x = nCoord -0.5;
    float y = 0;
    float w = (*n)->getChiSq();
    hfill_(id,x,y,w);
    w = 1;
    id = 7002;
    hfill_(id,x,y,w);
#endif
  }
#ifdef smoothChSq
  return sum;
#else
  int nPar = 4;
  if (nCoord>nPar) {
    return (sum * (1-nPar/nCoord));
  }
  else {
    return 0;
  }
#endif
}


double rangerTrack::totalRadThickness() const {
  
  double sum = 0;
  for (rangerNodeConstIterator n = begin(); n != end() ; n++) {
    sum += (*n)->radiationThickness();
  }
  return sum;
}

int rangerTrack::genCmp() const {
  int CMP = 0;
  for (rangerNodeConstIterator n = begin(); n != end() ; n++) {
    if ((*n)->hasHitb()) {
      if ((*n)->plane()>0) {
	ArtePointer<GEDE> ge((*n)->plane());
	float zsl = 0.5 * (((ArtePointer<GESL>) ge->gesl)->z1 
			   +((ArtePointer<GESL>) ge->gesl)->z2);

	if (               zsl <  190) CMP = CMP | Rsegc::vxd;  // set functional bits
	if (zsl >=  190 && zsl <  670) CMP = CMP | Rsegc::magt;
	if (zsl >=  670 && zsl < 1000) CMP = CMP | Rsegc::patt;
	if (zsl >= 1000 && zsl < 1330) CMP = CMP | Rsegc::bit6;
	if (zsl >= 1330 && zsl < 1400) CMP = CMP | Rsegc::ecal;
	if (zsl >= 1400              ) CMP = CMP | Rsegc::muon;

	switch (ge->cmp) {
	case 1:
	  CMP = CMP | Rsegc::sil;
	  break;
	case 2:
	  CMP = CMP | Rsegc::wch;
	  break;
	case 3:
	  CMP = CMP | Rsegc::msg;
	  break;
	case 4:
	  CMP = CMP | Rsegc::ptc;
	  break;
	case 5:
	  CMP = CMP | Rsegc::rich;
	  break;
	case 6:
	  CMP = CMP | Rsegc::trd;
	  break;
	case 7:
	  CMP = CMP | Rsegc::bit23;
	  break;
	case 8:
	  CMP = CMP | Rsegc::bit22;
	  break;
	default:
	  break;
	}
      }
    }
  }
  return CMP;
}


void rangerTrack::lastNoise() {
  (*last())->processNoise(tPartType);
}
