// 
//  Author:  Rainer Mankel                     Created:  29-Sep-98
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// 

// class rangerState
// base class for trajectory state

#ifndef rangerState_HH
#define rangerState_HH

#include "CLHEP/Matrix/SymMatrix.h"
#include "CLHEP/Matrix/Vector.h"
#include "arte/arte.hh"
#include "rang/ScatThick.hh"
#include "rang/FitMode.hh"

class rangerHitBase;
class rangerStateP;
class rangerStatePI;
class rangerStateL;
class rangerStateLI;
class rangerRadEl;

class rangerState {

public:
  rangerState();
  rangerState(ArtePointer<RHIT> rhit);
  rangerState(ArtePointer<HITB> hitb);
  rangerState(ArtePointer<RCCL> rccl);
  rangerState(const rangerState& s);  // copy constructor
  virtual ~rangerState();
  virtual rangerState& operator=(const rangerState& s) =0;  // assignment
  virtual rangerState& operator=(const rangerStateP& s) =0;  // assignment
  virtual rangerState& operator=(const rangerStatePI& ) {
    cout << "rangerState::operator=(const rangerStatePI& s)" << endl;
    exit(1);}
  virtual rangerState& operator=(const rangerStateL& ) {
    cout << "rangerState::operator=(const rangerStateL& s)" << endl;
    exit(1);}
  virtual rangerState& operator=(const rangerStateLI& ) {
    cout << "rangerState::operator=(const rangerStateLI& s)" << endl;
    exit(1);}

  inline const double& z() const { return zState;};
  inline const double& qPEst() const { return qPEstimate; }

  virtual const double x() const =0; // parameter lookup accessors
  virtual const double y() const =0;
  virtual const double tx() const =0;
  virtual const double ty() const =0;
  virtual const double qP() const { return 0; };
  virtual const double I() const { return 0; };

  inline double& setZ() { return zState;};      // update accessors
  inline double& setQpEst() {return qPEstimate; }

  virtual double& setX() =0; // parameter update accessors
  virtual double& setY() =0;
  virtual double& setTx() =0;
  virtual double& setTy() =0;
  virtual double& setQp() { cout << "State::setQp()" << endl; exit(1); };
  virtual double& setI() { cout << "State::setI()" << endl; exit(1); };

  virtual int dim() const =0;              // dimension of state vector
  virtual HepVector& vec() =0;             // lookup functions for state vector
  virtual const HepVector& vec() const =0; // 
  virtual HepSymMatrix& cov() =0;          //  lookup functions for cov matrix
  virtual const HepSymMatrix& cov() const =0; //

  static rangerState* makeState( const rangerFitMode fm);      
                                   // create empty state
  static rangerState* makeState(ArtePointer<RHIT>& rhit,
				const rangerFitMode fm); 
  static rangerState* makeState(ArtePointer<HITB>& hitb,
				const rangerFitMode fm); 
                                   // create from RHIT
  static rangerState* makeState(ArtePointer<RCCL>& rccl,int charge,
				const rangerFitMode fm); 
                                   // create from RCCL
  virtual rangerState* clone() const =0; // produce copy
  virtual void covInit()=0;
  virtual int predict(const double& zDestination,
	      const double& qPExp,rangerState*& targetState,
	      HepMatrix*& FTransport)=0;  // perform Kalman filter prediction
                                          // from this state to 
                                          // the target state
  virtual void project(HepVector& resid,HepMatrix& HH,rangerHitBase* hit)=0;
                                     // project into measurement space of hit 
  virtual void applyScat(const rangerScatThin& scat)=0;
                                        // apply m.s. correction 
  virtual void applyScat(const rangerScatThick& scat)=0;
                                        // apply m.s. correction 
  virtual void applyRadEl(const rangerRadEl& ){};
                                     // apply radiative energy loss correction 
  virtual void applyDeDx(const float ){};   
                                     // apply continuous energy loss correction
  virtual void applyRadPhoton(const float ,const float ){};   
                                     // apply resolved radiated photon
  virtual void storeRseg(ArtePointer<RSEG> rseg,int firstLast)=0; 
                                          // store track parameters into RSEG
  virtual void storeRtra(ArtePointer<RTRA> rseg,int firstLast)=0; 
                                          // store track parameters into RSEG
  virtual void storeRhit(ArtePointer<RHIT> rhit)=0; // store parameters as RHIT

  virtual ArtePointer<RPNT> storeRpnt(ArtePointer<RPNT> ){
    return 0; };                     // store parameters as RPNT
  void mergeRhit(ArtePointer<RHIT> rhit,float weight);
                                      // merge parameters to RHIT

  virtual void copyVec( const rangerState& s)=0; 
                                       // copy state vector from another state
                                       // (do not copy cov matrix)
  virtual void linExtrapVec( const rangerState& s, float z)=0;
                                       // extrapolate state vector from 
                                       //another state
  virtual rangerFitMode fitMode()=0;
  virtual void dump(){};               // dump private members to screen

protected:
  double zState;       // z of reference
  double qPEstimate;   // estimate for momentum parameter Q/p
};

#endif



