#include "rich/vessel.h"
#include "rich/Losses.h"

// ========================================================================
//
//    reflect: reflexion on a single planar mirror
//


int SinglePlaneMirror::reflect(Photon* p) {
  int ret;
  double path;
  if ((ret=intersection(*p,&path))) return ret;
  if ((ret=RICH_Vessel->trace(p,path))) return ret;
  
  p->n()-=double(2)*(ez()*p->n())*ez();  // short hand formula for the reflexion
  if (rnd_normal()>reflectivity().f(p->lambda())) return PMI_LOSS_REFLECTIVITY;
  return NO_LOSS;
  }

// -------------------------------------------------------------------------
//
//   reflexion used to find ring center
//


int SinglePlaneMirror::reflect(StraightLine* t) {
  int ret;
  double path;
  if ((ret=intersection(*t,&path))) return ret;

  t->x()+=path*t->n();
  t->n()-=double(2)*(ez()*t->n())*ez();  // short hand formula for the reflexion

  return NO_LOSS;
  }


// -----------------------------------------------------------------------
//
//  print data
//

void SinglePlaneMirror::print(FILE *fp) {
  fprintf(fp,"#\nSinglePlaneMirror\n");
//  fprintf(fp,"  reflectivity   %d\n",reflectivity());
  fprintf(fp,"  index       %d %d\n",row(),col());
  fprintf(fp,"  type        %s\n",type());
  fprintf(fp,"  mirrorname  %s\n",name());
//  fprintf(fp,"  end\n");
  Rectangle::print(fp,(char*)0);
  }

// -----------------------------------------------------------------------
//
//  initialisation from file
//

SinglePlaneMirror::SinglePlaneMirror(ScriptFile& f,Table& rrefl) {
  char n[256],tag[256];
  loss=-1;
  refl=rrefl;
  
  Vec3 xx0,vx,vy;
  int nnx=1,nny=1;
  double xcell=0,ycell=0;
  while (f.getline(n)) {
    char *s=n;
    sscanf(s,"%s",tag);s+=f.itemlength(s,1);
    int err=0;
    if (!strcmp(tag,"index"))   
      err=(sscanf(s,"%d %d",&ro,&co)!=1);  
    else if (!strcmp(tag,"type"))   
      err=(sscanf(s,"%s",&ttype[0])!=1);
    else if (!strcmp(tag,"mirrorname"))   
      err=(sscanf(s,"%s",&nname[0])!=1);  
    else if      (!strcmp(tag,"origin"))   
      err=(sscanf(s,"%lf %lf %lf",&xx0[0],&xx0[1],&xx0[2])!=3);
    else if (!strcmp(tag,"x_vector")) 
      err=(sscanf(s,"%lf %lf %lf",&vx[0],&vx[1],&vx[2])!=3);
    else if (!strcmp(tag,"y_vector"))   
      err=(sscanf(s,"%lf %lf %lf",&vy[0],&vy[1],&vy[2])!=3);
    else if (!strcmp(tag,"numcol")) {
      sscanf(s,"%d",&nnx);
      }
    else if (!strcmp(tag,"coldivision")) {
      sscanf(s,"%lf",&xcell);
      }
    else if (!strcmp(tag,"numrow")) {
      sscanf(s,"%d",&nny);
      }
    else if (!strcmp(tag,"rowdivision")) {
      sscanf(s,"%lf",&ycell);
      }
    else if (!strcmp(tag,"end"))  break;
    else {
      fprintf(stderr,
         "SinglePlaneMirror(ScriptFile):  Unknown tag: %s %d err=\n",tag,err);
      Exit(0);
      }
    }   
  Rectangle::operator=(Rectangle(xx0,vx,vy,nnx,xcell,nny,ycell));
   
  CSPM pm;
  pm.row=ro;
  pm.col=co;
  pm.orig[0]=x0().x();
  pm.orig[1]=x0().y();
  pm.orig[2]=x0().z();
  pm.xvec[0]=xsize()*ex().x();
  pm.xvec[1]=xsize()*ex().y();
  pm.xvec[2]=xsize()*ex().z();
  pm.yvec[0]=ysize()*ey().x();
  pm.yvec[1]=ysize()*ey().y();
  pm.yvec[2]=ysize()*ey().z();
  pm.iref=reflectivity().firstpoint();
  pm.nref=reflectivity().numpoints();

  memset(&pm.type[0],' ',sizeof(pm.type));
  int len=(strlen(ttype)>sizeof(pm.type)) ? sizeof(pm.type) :strlen(ttype);
  memcpy(&pm.type[0],ttype,len);

  memset(&pm.name[0],' ',sizeof(pm.name));
  len=(strlen(nname)>sizeof(pm.name)) ? sizeof(pm.name) :strlen(nname);
  memcpy(&pm.name[0],nname,len);
  
  ArteTable<CSPM>::push_back(pm);

  }


SinglePlaneMirror::SinglePlaneMirror(ArtePointer<CSPM> a)
  : Rectangle(Vec3(a->orig),Vec3(a->xvec),Vec3(a->yvec)),
    refl(a->iref,a->nref) {
  int i;
  
  ro=a->row;
  co=a->col;
  
  memcpy(ttype,a->type,sizeof(a->type));
  for (i=sizeof(a->type)-1;i>=0;i--) if (ttype[i]!=' ') break;
  ttype[i+1]='\0';
  
  memcpy(nname,a->name,sizeof(a->name));
  for (i=sizeof(a->name)-1;i>=0;i--) if (nname[i]!=' ') break;
  nname[i+1]='\0';
  }

SinglePlaneMirror::SinglePlaneMirror(const SinglePlaneMirror& s) {
  memcpy(this,&s,sizeof(SinglePlaneMirror));
  }
