// -*-C++-*-
#ifndef _RICHEVENTMC_H_
#define _RICHEVENTMC_H_

#include <map.h>
#include <multimap.h>
#include <math.h>
#include "arte/arte.hh"
#include "rich/ricutil.h"
#include "CLHEP/Hist/HBookTuple.h"

#ifdef USE_ROOT
#include "TMinuit.h"
#endif

// -----------------------  RICH photons ----------------------------

class RICHphotons {
 private:
  // mtra=key,  mimp=line in MIMP table
  multimap <unsigned const, ArtePointer<MIMP>, less<unsigned> > photonList;
  multimap <unsigned const, ArtePointer<MIMP>, less<unsigned> >::iterator it;
  multimap <unsigned const, ArtePointer<MIMP>, less<unsigned> >::iterator itl;

 public:
  RICHphotons();

  // --- iterations
  inline void   begin() { 
    it =photonList.begin();
    itl=photonList.end();
  }
  inline void   begin(unsigned imtra) {
    it =photonList.lower_bound(imtra);
    itl=photonList.upper_bound(imtra);
  }
  inline void   next()  { ++it;           }
  inline bool   valid() { return it!=itl; }
  inline bool   end()   { return it==photonList.end(); }
  inline bool   find(unsigned imtra) {
    if ((*it).first!=imtra) {
      it=photonList.find(imtra);
      return it!=photonList.end();
    }
    return 1;
  }

  // --- data access
  inline unsigned          mtra() { return (*it).first; }
  inline ArtePointer<MIMP> mimp() { return (*it).second; }
  inline int id() { 
    int id=-1;
    for (ArteRelationIterator<DIGB>::iterator idigb=mimp()->digb.begin();
	 idigb!=mimp()->digb.end(); ++idigb) id=(*idigb)->id;
    return id;
    }
  inline bool detected() { return id() > -1; }
  inline float DetectedPhi() { 
    ArteRelationIterator<HITB>::iterator ihitb=mimp()->hitb.begin();
    return (*ihitb)->q;
  } 
  inline ArtePointer<HITB> hitb() { return (*mimp()->hitb.begin()); }
  inline float DetectedLambda() { 
    ArteRelationIterator<HITB>::iterator ihitb=mimp()->hitb.begin();
    return (*ihitb)->r;
  }  
  void print();
  void print(unsigned imtra);
};




// ----------------------- MC RICH rings ----------------------------

struct RichRingData {
  Vec3  x;         // point in RICH
  float nrefr;     // refraction index
  float tx,ty;     // direction
  float radius;    // radius
  int   hitbs;     // detected photons
  int   mimps;     // MC radiated photons (after reduction)
  ArtePointer<TOPP> topp;   // type
  ArtePointer<TOPP> mother; // mother type
  ArtePointer<MVER> vertex; // vertex
};

class RICHrings {
 private:
  // mtra=key
  map <unsigned const, RichRingData, less<unsigned> > ringList;
  map <unsigned const, RichRingData, less<unsigned> >::iterator it;
  map <unsigned const, RichRingData, less<unsigned> >::iterator itl;
  RICHphotons  *photons;
  RichRingData data;

 public:
  /// Constructor
  RICHrings();
  /// Container
  inline void   begin() { 
    it =ringList.begin();
    itl=ringList.end();
  }
  inline void next()  { ++it;           }
  inline bool valid() { return it!=itl; }
  inline bool find(unsigned imtra) { 
    if ((*it).first!=imtra) {
      it=ringList.find(imtra);
      return it!=ringList.end();
    }
    return 1;
  }

  /// Print rings
  void print();
  void printPhotons() { photons->print(); }
  void fill(HBookTuple *tuple);

  /// Data
  inline ArtePointer<MTRA> mtra() { return ArtePointer<MTRA>((*it).first); }
  inline float RefractionIndex()  { return (*it).second.nrefr; }
  inline Vec3  x()       { return (*it).second.x; }
  inline float tx()      { return (*it).second.tx; }
  inline float ty()      { return (*it).second.ty; }
  inline float phi()     { return atan(tx()); }
  inline float lambda()  { return atan(ty()*cos(phi())); }
  inline float radius()  { return (*it).second.radius; }
  inline int   DetectedPhotons()  { return (*it).second.hitbs; }
  inline int   RadiatedPhotons()  { return (*it).second.mimps; }
  inline void  firstPhoton(unsigned imtra) { photons->begin(imtra); }
  inline bool  validPhoton() { return photons->valid(); }
  inline void  nextPhoton()  { photons->next(); }
  float  momentum();
  inline ArtePointer<TOPP> topp()   { return (*it).second.topp; }
  inline ArtePointer<TOPP> mother() { return (*it).second.mother; }
  inline ArtePointer<MVER> vertex() { return (*it).second.vertex; }
  inline RICHphotons *GetPhotons()   { return photons; } 
  inline Vec3 T() {
    Vec3   x0=x();
    Vec3   n0=normalize(Vec3(tx(),ty(),1));
    Vec3   x1=x0-n0*x0.z()/n0.z();
    float  delta=norm(cross(x1,x0))/norm(x1-x0);
    float  t=sqrt(norm(x0)*norm(x0)-delta*delta);
    return x0-t*n0;
  }
};




// ----------------------- RC RICH rings ----------------------------

struct RichRecoData {
  float tx,ty;     // direction
  float radius;    // radius
  float photons;   // reconstructed photons
};

struct RecoKey {
  unsigned mtra;
  float dtx,dty;   // direction precision
  float dradius;   // radius precision
  float dphotons;  // photon precision
};

struct RichRule {
  bool operator()(const RecoKey k1,const RecoKey k2) const {
    if (k1.mtra==k2.mtra) {
      return (k1.dtx*k1.dtx+k1.dty*k1.dty <
              k2.dtx*k2.dtx+k2.dty*k2.dty);
    }
    else return (k1.mtra<k2.mtra);
  }
};

class RICHreco {
 private:
  // mtra=key
  multimap < RecoKey const, RichRecoData, RichRule > recoList;
  multimap < RecoKey const, RichRecoData, RichRule >::iterator it;
  multimap < RecoKey const, RichRecoData, RichRule >::iterator itl;
  RichRecoData data;
  RecoKey      key;
  RICHrings    *mcRings;

 public:
  /// Constructor
  RICHreco();
  /// Container
  inline void   begin() { 
    it =recoList.begin();
    itl=recoList.end();
  }
  inline void   next()  { ++it;           }
  inline bool   valid() { return it!=itl; }


  /// Print rings
  void print();
  void fill(HBookTuple *tuple);


  inline ArtePointer<MTRA> mtra() { return ArtePointer<MTRA>((*it).first.mtra); }
  inline float tx()      { return (*it).second.tx; }
  inline float dtx()     { return (*it).first.dtx; }
  inline float ty()      { return (*it).second.ty; }
  inline float dty()     { return (*it).first.dty; }
  inline float phi()     { return atan(tx()); }
  inline float lambda()  { return atan(ty()*cos(phi())); }
  inline float radius()  { return (*it).second.radius; }
  inline float dradius() { return (*it).first.dradius; }
  inline float photons() { return (*it).second.photons; }
  inline float dphotons(){ return (*it).first.dphotons; }
};




// ----------------------- RICH angles --------------------------------

#ifdef USE_ROOT

class RICHangles {
 private:
  RICHrings *mcRings;
 public:
  RICHangles();
  void fitRings(HBookTuple *tuple);
};
#endif


#endif











