#ifdef USE_ROOT
#include "OTrEfficiencyH1F.hh"

//__________________________________________________________________
//
// OTrEfficiencyH1F
//
// An OTrEfficiencyH1F is a TH1F especially designed for analyzing efficiency as a function of one parameter, e.g. the distance to a wire. 
//
// Author: Wouter Hulsbergen 
//__________________________________________________________________

ClassImp(OTrEfficiencyH1F) ;
//void OTrEfficiencyH1F::Streamer(TBuffer &) {} ;
//void OTrEfficiencyH1F::ShowMembers(TMemberInspector &, char *) {} ;

OTrEfficiencyH1F::OTrEfficiencyH1F(const Text_t *const name, Text_t* title, 
				   Int_t nbinsx, Axis_t xlow, const Axis_t xup)
  : TH1F(name,title,nbinsx,xlow,xup)
{
  fNfound = new float[nbinsx+2];
  fNall   = new float[nbinsx+2];
  Reset() ;
}

OTrEfficiencyH1F::~OTrEfficiencyH1F()
{
  delete[] fNfound ;
  delete[] fNall ;
} ;

void OTrEfficiencyH1F::Reset()
{
  TH1F::Reset() ;
  for(int binx=0; binx<=GetNbinsX()+1; binx++)
    fNfound[binx] = fNall[binx] = 0;
  Calculate() ;
}

void OTrEfficiencyH1F::Fill(Axis_t x, bool found, Stat_t w)
{
  // Fill statement. Does not call TH1F::Fill. 
  Int_t bin = GetXaxis()->FindBin(x);
  fNall[bin] += w;
  if(found) fNfound[bin] += w;
}

void OTrEfficiencyH1F::Calculate()
{
  // Fills the entries of the TH1F x-axis array. Only after calling this function the TH1F is filled with the efficiencies.
  for(int binx=0; binx<=GetNbinsX()+1; binx++)
    {
      int bin = GetBin(binx) ;
      float N = fNall[bin] ;
      float n = fNfound[bin] ;
      SetBinContent(bin,(n+1)/float(N+2)) ;
      SetBinError(bin,sqrt(double(n+1)*double(N+1-n)/
			   (double(N+3)*double(N+2)*double(N+2)))) ;
    }
}

void OTrEfficiencyH1F::Draw(Option_t* option)
{
  Calculate();
  TH1F::Draw(option);
};

void OTrEfficiencyH1F::Paint(Option_t* option)
{
  Calculate();
  TH1F::Paint(option);
};

void OTrEfficiencyH1F::PrintEmptyBins()
{
  cout << "Empty bins in " << GetName() << ":" ;
  for(int binx=0; binx<GetNbinsX(); binx++)
    if(fNall[binx]==0) cout << binx << " " ;
  cout << endl ;
}

void OTrEfficiencyH1F::Print()
{
  Calculate();
  cout << GetName() << ":Print()" << endl;
  for(int binx=0; binx<=GetNbinsX()+1; binx++)
    {
      cout << binx << " " << fNall[binx] << " "
	   << fNfound[binx] << " " 
	   << GetBinContent(GetBin(binx)) << " "
	   << GetBinError(GetBin(binx)) << endl;
    }
}

#endif //USE_ROOT











