#include "OTrNodeProjection.hh"
#include "OTrReconstructedHit.hh"
#include "OTrTrackDescription.hh"
#include "OTrParameterizedRTRelation.hh"
#include "OTrSection.hh"
#include "OTrGede.hh"
#include "OTrHit.hh"

#define TRACKFITT
#undef TRACKFITT

//__________________________________________________________________
//
// OTrNodeProjection
//
// This class stores some values which have to do with the
// trackfit. It is only meant to improve speed. All its data members
// can be calculated from an OTrHit or OTrReconstructedHit and a
// corresponding OTrTrack.
//
// Author: Wouter Hulsbergen 
//__________________________________________________________________

#ifdef USE_ROOT
ClassImp(OTrNodeProjection) ;
//void OTrNodeProjection::Streamer(TBuffer &) {} ;
//void OTrNodeProjection::ShowMembers(TMemberInspector &, char *) {} ;
#endif

inline void
OTrNodeProjection::InitGeometry(const OTrWire* wire)
{
  //double theta = -1*wire->GetGede()->GetStereoAngle() ;
  fZw = wire->GetRawZ() ;
  fProjVector[0] =  wire->GetGede()->GetCosStereoAngle() ; //cos(theta) ;
  fProjVector[1] = -wire->GetGede()->GetSinStereoAngle() ; //sin(theta) ;
  fProjVector[2] = fProjVector[0]*fZw ;
  fProjVector[3] = fProjVector[1]*fZw ;
  fXw = wire->GetRawX() ;
  fRTRelation = wire->GetChannel()->GetRTRelation() ;
  fT0 = fRTRelation->GetT(0).GetValue() ;
  fSigmaC2 = sqr(wire->GetGede()->GetWirePitch())/12 ;
#ifdef TRACKFITT
  fSigmaT2 = fRTRelation->GetSigma2(fRm).GetValue() ;
  fSigmaR2 = fSigmaT2 / sqr(fRTRelation->GetT(fRm).ddx) ;
#else
  fSigmaR2 = fRTRelation->GetSigma2(fRm).GetValue() ;
  fSigmaT2 = fSigmaR2 * sqr(fRTRelation->GetT(fRm).ddx) ;
#endif
}

inline void
OTrNodeProjection::Init(const OTrHit* measuredhit)
{
  InitGeometry(measuredhit->GetWire()) ;
  fTm = measuredhit->GetDriftTime() ;
  fRm = fRTRelation->GetR(fTm) ;
}
  
OTrNodeProjection::OTrNodeProjection(OTrReconstructedHit* hit) 
  : fReconstructedHit(hit), fProjVector(4), fdChi2dQ(), fd2Chi2dQ2()
{
  Init(fReconstructedHit->GetMeasuredHit()) ;
} ;

OTrNodeProjection::OTrNodeProjection(const OTrHit* hit) 
  : fReconstructedHit(0), fProjVector(4), fdChi2dQ(), fd2Chi2dQ2()
{
  Init(hit) ;
} ;

OTrNodeProjection::OTrNodeProjection(const OTrWire* wire)
  : fReconstructedHit(0), fProjVector(4), fdChi2dQ(), fd2Chi2dQ2()
{
  InitGeometry(wire) ;
  fTm = fT0 ;
  fRm = 0 ;
}

OTrNodeProjection::~OTrNodeProjection() {};

void OTrNodeProjection::Project(const OTrTrackDescription* track, bool updateInternal, 
				bool calcDerivatives)
{
  float tprojection = 
    track->GetTx()*fProjVector[0] + track->GetTy()*fProjVector[1] ;
  float noemer2 = 1 + sqr(tprojection) ;
  float noemer = sqrt( noemer2 ) ;
  
  int dim = track->GetDim() ;
  // I have added the z-coordinate where (x,y) is calculated, to simplify
  // the addition of a target constraint different from (0,0,0) ;
  // fD = ((track->GetX()-track->GetTx()*track->GetZ()) *fProjVector[0]+
  // 	(track->GetY()-track->GetTy()*track->GetZ()) *fProjVector[1]+
  // 	track->GetTx()*fProjVector[2]+trackGetTy()*fProjVector[3]
  // 	-fXw) / noemer ;
  
  if(dim==4) fD = ( dot(track->GetQ(),fProjVector) - fXw) / noemer ;
  else fD = ( track->GetTx()*fProjVector[2] + track->GetTy()*fProjVector[3] - fXw) / noemer ; 
  
  if(calcDerivatives) {
    if(fdDdQ.num_row()!=dim) {
      fdDdQ = HepVector(dim) ;
      //fdChi2dQ = HepVector(dim) ;
      fd2Chi2dQ2 = HepSymMatrix(dim) ;
    };
    
    if(dim==4) {
      //fD = ( dot(track->GetQ(),fProjVector) - fXw) / noemer ;
      fdDdQ = fProjVector/noemer ;
      // Are these negligable ?!
      fdDdQ[2] += fD * -fProjVector[0]*tprojection/noemer2 ;
      fdDdQ[3] += fD * -fProjVector[1]*tprojection/noemer2 ;
    } else {
      //fD = ( track->GetTx()*fProjVector[2] + track->GetTy()*fProjVector[3] - fXw) / noemer ; 
      //fdDdQ[0] = fProjVector[2]/noemer + fD * -fProjVector[0]*tprojection/noemer2 ;
      // fdDdQ[1] = fProjVector[3]/noemer + fD * -fProjVector[1]*tprojection/noemer2 ;
      fdDdQ[0] = fProjVector[2]/noemer ;
      fdDdQ[1] = fProjVector[3]/noemer ;
    };
    
    int signD = 1; if(fD<0.) signD=-1;
    int lrsign = signD;
    if(fReconstructedHit) {
      if(updateInternal && !fReconstructedHit->GetLRSignIsFixed()) 
	fReconstructedHit->SetLRSign(fD) ;
      lrsign = fReconstructedHit->GetLRSign() ;
      if(!fReconstructedHit->GetLRSignIsFixed() && lrsign*signD <0) fEtaError = true ;
      else fEtaError = false ;
    };
    
    // Calculate residual
    fT = fRTRelation->GetT(signD*fD) ;
    fResidualT =  (fT.GetValue() - fT0) - lrsign*signD*( fTm - fT0 ) ;
    fResidualR =  (signD*fD) - (lrsign*signD*fRm) ;
    fResidualD = fD - lrsign*fRm ;
    
    float weight = 1/fSigmaR2 ;
    if(lrsign==0) weight = 1/fSigmaC2 ;
    
    //   bool fitTimes = false ;
    //   // Calculate Chi2 derivatives
    //   if(fitTimes) {
    //     fChi2 = sqr(fResidualT)/fSigmaT2 ;
    //     fdChi2dQ = 2*fResidualT/fSigmaT2 * signD*fT.ddx*fdDdQ ;  
    //     for(int row=0; row<dim; row++)
    //       for(int col=0; col<=row; col++)
    // 	fd2Chi2dQ2[row][col] = 
    // 	  2/fSigmaT2 *fdDdQ[row]*fdDdQ[col] * (fT.ddx*fT.ddx + fResidualT* fT.d2dx2) ;
    //   } else { 
    fChi2 = weight*sqr(fResidualD) ;
    fdChi2dQ = 2*weight*fResidualD * fdDdQ ; //fResidualR*signD  
    for(int row=0; row<dim; row++)
      for(int col=0; col<=row; col++)
	fd2Chi2dQ2[row][col] = 2*weight * fdDdQ[row]*fdDdQ[col] ;
    //};
  };
};

void
OTrNodeProjection::Dump() const 
{
  cout << "fD: " << fD << " fRm: " << fRm << " fTm: " << fTm << endl
       << "fT: " << fT << endl
       << "fResidual: " << fResidualT << " " << fResidualR << endl 
       << "fSigma2: " << fSigmaT2 << " " << fSigmaR2 << endl 
       << "fChi2: " << fChi2 << endl
       << "fdChi2dQ: " << fdChi2dQ
       << "fdDdQ: " << fdDdQ << endl ;
};

