#include "OTrSplineRTRelation.hh"
#include <math.h>
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>

#ifdef USE_ROOT
 #include <TH1.h>
#endif

//const ParameterizedRTRelationType OTrSplineRTRelation::fgType = spline;

OTrSplineRTRelation::~OTrSplineRTRelation() {};

vector<OTrFunctionValue>
OTrSplineRTRelation::CalcBasisFuncVector(const double x, const int size) const
{
  vector<OTrFunctionValue> bfv(size);
  const HepVector* knots ;
  int degree ;
  if(size==GetNoTParameters()) {
    knots  = &fTKnots ;
    degree = fTDegree ;
  } else if (size==GetNoSigma2Parameters()) {
    knots = &fSigma2Knots ;
    degree = fSigma2Degree ;
  } else {
    cout << "SplineRTRelation::CalcBasisFuncVector error !!!!!" << endl;
    return bfv;
  }
  int i;
  for(i=0; i<size; i++)
    {
      bfv[i].value = EvaluateBSpline(*knots, i, degree, x) ;
      bfv[i].ddx   = EvaluateBSplineDerivative(*knots, i, degree, x) ;
      bfv[i].d2dx2 = EvaluateBSplineSecondDerivative(*knots, i, degree, x) ;
      if(fabs(bfv[i].value)<1e-16) bfv[i].value=0;
      if(fabs(bfv[i].ddx)<1e-16) bfv[i].ddx=0;
      if(fabs(bfv[i].d2dx2)<1e-16) bfv[i].d2dx2=0;
    }
  return bfv ;
};

void
OTrSplineRTRelation::CorrectT0()
{
  cout << "SplineRTRelation::CorrectT0()" ;
  HepVector TParameters = GetTParameters() ;
  HepDouble deltaT0 = CalcFunctionT(GetRMin()).value ;
  HepDouble newT0 = GetT0() + deltaT0 ;
  SetT0(newT0) ;
  for(int i=0; i<GetNoTParameters(); i++)
    TParameters[i] -=  deltaT0 ;
  cout << " (old,delta,new) " 
       << newT0-deltaT0 << " " << deltaT0 << " " << newT0 << endl ;
  SetTParameters(TParameters) ;
  return;
};

void OTrSplineRTRelation::RepairTParameters()
{
  if(sCorrectionLevel<=0) return;
  HepVector TParameters = GetTParameters() ;
  int noParameters = GetNoTParameters();
  int middle = noParameters/2;
  int i;
  cout << "SplineRTRelation::RepairTParameters() repaired: " ;
  for(i=0; i<middle; i++)
    if(TParameters[i]<TParameters[i+1]) {
      if(i>0) TParameters[i]=0.5*(TParameters[i-1]+TParameters[i+1]) ;
      else TParameters[i]=TParameters[i+1]+0.5*(TParameters[i+1]-TParameters[i+2]);
      cout << i << " " ;
    };
  for(i=middle; i<noParameters; i++)
    if(TParameters[i]>TParameters[i-1]) {
      if(i+1<noParameters) TParameters[i]=0.5*(TParameters[i-1]+TParameters[i+1]) ;
      else TParameters[i]=TParameters[i-1]+0.5*(TParameters[i-1]-TParameters[i-2]) ;
      cout << i << " " ;
    };
  cout << endl;
  SetTParameters(TParameters) ;
  return;
};

void
OTrSplineRTRelation::RepairSigma2Parameters()
{
  if(sCorrectionLevel<=0) return;

  HepVector Sigma2Parameters = GetSigma2Parameters() ;
  int noSigma2Parameters = GetNoSigma2Parameters();
  if(sCorrectionLevel>=2 && Sigma2Parameters[0]<Sigma2Parameters[1]/5)
    Sigma2Parameters[0]=Sigma2Parameters[1]/5 ;
  if(Sigma2Parameters[noSigma2Parameters-1]<Sigma2Parameters[0])
      Sigma2Parameters[noSigma2Parameters-1]=Sigma2Parameters[0] ;
  SetSigma2Parameters(Sigma2Parameters);
  return;
};

  
void
OTrSplineRTRelation::SetUniformSigma(HepDouble sigma)
{
  int noSigma2Parameters = GetNoSigma2Parameters();
  HepDouble sigma2 = sigma*sigma;
  HepVector    sigma2Parameters(noSigma2Parameters,0) ;
  for(int i=0; i<noSigma2Parameters; i++)
    sigma2Parameters[i] = sigma2 ;
  HepSymMatrix sigma2Cov(noSigma2Parameters,0) ;
  UpdateSigma2Parameters(sigma2Parameters,sigma2Cov) ;
};

void
OTrSplineRTRelation::CalcEquidistantKnots()
{
  // knots for T
  int k = fTDegree;
  int m = GetNoTParameters() + k + 1 ;
  HepDouble deltaXi = 2./(m - 2*k - 1) ;
  HepVector tKnots(m) ;

  int i;
  for(i=0; i<k+1; i++)
    tKnots[i] = -1. ;
  for(i=k+1; i<m-k-1; i++)
    tKnots[i] = -1. + (i - k) * deltaXi ;
  for(i=m-k-1; i<m; i++)
    tKnots[i] = 1. ;
  fTKnots = tKnots ;
  //cout << "TKnots: " << fTKnots ;

  // knots for sigma2
  k = fSigma2Degree ;
  m = GetNoSigma2Parameters() + k + 1 ;
  deltaXi = 2./(m - 2*k - 1) ;
  HepVector sigma2Knots(m) ;
  for(i=0; i<k+1; i++)
    sigma2Knots[i] = -1. ;
  for(i=k+1; i<m-k-1; i++)
    sigma2Knots[i] = -1. + (i - k) * deltaXi ;
  for(i=m-k-1; i<m; i++)
    sigma2Knots[i] = 1. ;
  fSigma2Knots = sigma2Knots ;
  //cout << "Sigma2Knots: " << fSigma2Knots ;
};

void
OTrSplineRTRelation::Write(ostream& ofs) const {
  OTrParameterizedRTRelation::Write(ofs) ;
  ofs << endl << fTDegree << " " << fSigma2Degree << endl ;
}

OTrSplineRTRelation::OTrSplineRTRelation(istream& ifs) :
  OTrParameterizedRTRelation()
{
  OTrParameterizedRTRelation::Read(ifs) ;
  ifs >> fTDegree >> fSigma2Degree ;
  CalcEquidistantKnots() ;
  FillBasisFuncTables() ;
  UpdateTParameters() ;
  UpdateSigma2Parameters();
} ;


#ifdef USE_ROOT
OTrSplineRTRelation::OTrSplineRTRelation(int noTParameters,
					 int noSigma2Parameters,
					 TH1* timeHisto,
					 HepDouble rmin, HepDouble rmax,
					 HepDouble estimated_time_Rmin,
					 HepDouble estimated_time_Rmax,
					 HepDouble estimated_time_error,
					 int tdegree, int s2degree) :
  OTrParameterizedRTRelation(noTParameters, noSigma2Parameters, rmin, rmax), 
  fTDegree(tdegree), fSigma2Degree(s2degree) 
{
  cout << "SplineRTRelation::SplineRTRelation(...)" << endl ;

  //Set time boundaries
  double tMin = estimated_time_Rmin ;
  double tMax = estimated_time_Rmax ;
  int minBin = timeHisto->GetXaxis()->FindBin(tMin) ;
  int maxBin = timeHisto->GetXaxis()->FindBin(tMax) ;
  if(minBin>maxBin) {
    int tmp=minBin; minBin=maxBin; maxBin=tmp ; }
  const int noBins = maxBin-minBin + 1 ;
  
  //Integrate spectrum
  vector<double> integratedspectrum(noBins) ;
  for(int j=0; j<noBins; j++) integratedspectrum[j]=0;
  for(int i=0; i<noBins; i++)
    for(int j=i; j<noBins; j++)
      integratedspectrum[j]+= timeHisto->GetBinContent(i+minBin) ;

//   for(int j=0; j<noBins; j++)
//     cout << j << " " << integratedspectrum[j] << endl ;

  double noEntries = integratedspectrum[noBins-1] ;
  // Define knots
  CalcEquidistantKnots() ;

  //Define Schoenberg's knots
  HepVector schoenbergknots(noTParameters) ;
  if(fTDegree==0)
    for(int i=0; i< noTParameters; i++)
      schoenbergknots[i] = (0.5*(fTKnots[i]+fTKnots[i+1])) ;
  else
    for(int i=0; i< noTParameters ; i++) {
      double sum(0.) ;
      for(int j = i+1 ; j < i + fTDegree + 1; j++) sum += fTKnots[j];
      schoenbergknots[i] = sum/(double)fTDegree ;
    };
  
  //Set initial parameters by spec value on Schoenberg's knots
  HepDouble norm = 2./noEntries ; // effectiveCellSize/GetRMax() ; 
  HepVector TParameters(noTParameters) ;
  for(int j=0; j< noTParameters ; j++) {
    TParameters[j] = tMax ;
    for(int i=0; i< noBins ; i++) {
//    cout << i << " " << tMin << " " << tMax << " " 
// 	   << (noEntries-integratedspectrum[i])*norm << " "
// 	   << schoenbergknots[j]+1 << endl ;
      if(((tMax>tMin) && (integratedspectrum[i]*norm>=schoenbergknots[j]+1)) ||
	 ((tMin>tMax) && ((noEntries-integratedspectrum[i])*norm<=schoenbergknots[j]+1)))
	{
	  TParameters[j] = timeHisto->GetBinCenter(i+minBin) ;
	  break ;
	};
    };
  };
  
  UpdateTParameters(TParameters) ;
  cout << "Initial TParameters: " << GetTParameters() ;
  
  //Set initial sigma parameters
  SetUniformSigma(estimated_time_error) ;
  cout << "Initial Sigma2Parameters: " << GetSigma2Parameters() ;
  //Shit, dit is heel lelijk
  FillBasisFuncTables() ;
};
#endif


#define FUF 1e-16

//Evaluate j'th basis spline of degree k in x on knots t   
double
OTrSplineRTRelation::EvaluateBSpline(const HepVector& t,
				  const int j, const int k, 
				  const double x) const
{
  //cout << "SplineRTRelation::EvaluateBSpline()" << endl;
  if(j < 0 || j+k+1 >= t.num_row() ) return 0 ;
  if(t[j+k+1] <= x || t[j] > x) return 0 ;

  if(k==0) return 1 ;
  
  double v ;
  if(fabs(t[j+k]-t[j])< FUF) v = 0 ;
  else v = (x-t[j])/(t[k+j] - t[j]) * EvaluateBSpline(t, j, k-1, x) ;

  double w ;
  if(fabs(t[j+k+1]-t[j+1])< FUF) w = 0 ;
  else w = (t[j+k+1]-x)/(t[j+k+1]-t[j+1]) * EvaluateBSpline(t, j+1, k-1, x) ;
  
  return v + w ;
};

double
OTrSplineRTRelation::EvaluateBSplineDerivative(const HepVector& t,
					    const int j, const int k, 
					    const double x) const
{
  if(k==0) return 0 ;
  
  double vbar ;
  if(fabs(t[j+k]-t[j])< FUF) vbar = 0 ;
  else vbar = 1./(t[k+j] - t[j]) * EvaluateBSpline(t, j, k-1, x) ;

  double wbar ;
  if(fabs(t[j+k+1]-t[j+1])< FUF) wbar = 0 ;
  else wbar = 1./(t[j+k+1]-t[j+1]) * EvaluateBSpline(t, j+1, k-1, x) ; 
  return k*(vbar - wbar) ;
}


double
OTrSplineRTRelation::EvaluateBSplineSecondDerivative(const HepVector& t,
						  const int j, const int k, 
						  const double x) const
{
  if(k<=1) return 0 ;
  
  double vhat ;
  if(fabs(t[j+k]-t[j])< FUF) vhat = 0 ;
  else vhat = 1./(t[k+j] - t[j]) * 
         EvaluateBSplineDerivative(t, j, k-1, x) ;

  double what ;
  if(fabs(t[j+k+1]-t[j+1])< FUF) what = 0 ;
  else what = 1./(t[j+k+1]-t[j+1]) * 
         EvaluateBSplineDerivative(t, j+1, k-1, x) ; 
  return k*(vhat - what) ;
}

#undef FUF









