#ifdef USE_ROOT
#include "OTrTdcHistogram.hh"
#include "OTrChannel.hh"
#include "OTrSector.hh"
#include "OTrObservable.hh"
#include <TMinuit.h>

OTrTdcHistogram::OTrTdcHistogram() :
  TH1D()
{
}

OTrTdcHistogram::OTrTdcHistogram( OTrChannel* c ) :
  TH1D( c->GetTimeH1() )
{
}

OTrTdcHistogram::OTrTdcHistogram( OTrSector* s ) :
  TH1D( s->GetTimeH1() )
{
}

//
// An asymmetric Gaussian to describe the shape of the TDC spectrum:
//
double asymmetricGaussian(double* x, double* par )
{
  double argu;
  if ( x[0] > par[0] ) {
    argu = (x[0]-par[0])/par[1]; 
  } else {
    argu = (x[0]-par[0])/par[2]; 
  }
  double gausf = exp( -0.5*argu*argu);
  return par[4]*((1-par[3])*gausf + par[3]);
}

//
// Sum of two error functions to describe the shape of the TDC spectrum:
//
double twoErrorFunctions(double* x, double* par)
{
  double h1 = (x[0]-par[0])/par[1]/sqrt(2);
  double h2 = (x[0]-par[2])/par[3]/sqrt(2);
  double h = (erfc(h2)+erfc(h1)-2);
    // /fabs(h2-h1);
  return par[4]*(h + par[5]);
}



int OTrTdcHistogram::MaximumBin() 
{	
  double maxCount = GetBinContent(1) ;
  int maxBin = 1; 
  for ( int bin=2; bin < GetNbinsX(); bin++) {
    double bincontent = GetBinContent(bin);
    if ( bincontent > maxCount ) {
      maxCount = bincontent;
      maxBin = bin;
    }
  }
  return maxBin;
}

OTrFObservable OTrTdcHistogram::GetPeakValue( bool withGaussianFit )
{
  float T0 = 0;
  float T0error = 1;
  float center = MaximumX();

  cout << "center = " << center 
           << " nr of bins : " << GetNbinsX() << endl;

  if ( !withGaussianFit ) {
    T0 = -center;
  } else {
    TF1* f1 = new TF1("eengausje","gaus") ;
    f1->SetParameter(0, GetBinContent(MaximumBin()) ) ;
    f1->SetParameter(1,center) ;
    f1->SetParameter(2,2) ;
    f1->SetRange(center-7,center+7) ;
    f1->SetParLimits(1,center-5,center+5) ;
    //      Fit("gausje","QR0") ; 
    Fit("eengausje","RB") ;
    if( fabs(f1->GetParameter(1)-center) < 4 ) {
      T0 =  - f1->GetParameter(1) ;
      T0error = f1->GetParError(1) ;
    } else {
      cout << "WARNING: " 
	   << f1->GetParameter(1)-center << endl ;
    }
  } 
  return OTrFObservable( T0, T0error );
}

OTrFObservable OTrTdcHistogram::FitAsymmetricGaussian( OTrFObservable& width )
{
  float T0 = 0;
  float T0error = 1;

  int nFitPar = 5;
  TF1* f1 = new TF1("f1", asymmetricGaussian, 0, 256, nFitPar ) ;
  if ( gMinuit ) gMinuit->SetMaxIterations( 2000 );
  double startpar[nFitPar];
  startpar[0] = MaximumX();
  startpar[1] = 20; 
  startpar[2] = 20; 
  startpar[3] = 0.02;
  startpar[4] = 1;
  for ( int i = 0; i < nFitPar; i++ ) f1->SetParameter(i,startpar[i]);
  f1->SetParLimits(0,20,250) ;
  f1->SetParLimits(1,5,50) ;
  f1->SetParLimits(2,5,100) ;
  f1->SetParLimits(3,0,0.7) ;
  f1->SetRange(20,251) ;

  Fit("f1","RBQ0") ;
  cout << "par 0: " << f1->GetParameter(0) << " " 
       << "par 1: " << f1->GetParameter(1) << " "
       << "par 2: " << f1->GetParameter(2) << " "
       << "(par 1 + par2)/2 = " << (f1->GetParameter(1)+f1->GetParameter(2))/2
       << endl;
  Ssiz_t indx = gMinuit->fCstatu.Index("CONVERGED");
  //cout << "index = " << indx << endl;
  if ( indx >= 0 ) {
    T0 = -f1->GetParameter(0) - 2*f1->GetParameter(1);
    T0error = f1->GetParError(0);
    float val = f1->GetParameter(1) + f1->GetParameter(2);
    float eval = sqrt(f1->GetParError(1)*f1->GetParError(1) +
                      f1->GetParError(2)*f1->GetParError(2));
    width = OTrFObservable( val, eval );
  } else {
    cout << "** fit did not converge." << endl;
    width = OTrFObservable(0,0);
  }
  cout << "T0, T0error = " << T0 << " " << T0error << endl;
  delete f1;
  return OTrFObservable( T0, T0error );
}

OTrFObservable OTrTdcHistogram::FitTwoErrorFunctions()
{
  float T0 = 0;
  float T0error = 1;

  int nFitPar = 6;
  TF1* f1 = new TF1("f1", twoErrorFunctions, 0, 256, nFitPar ) ;
  if ( gMinuit ) gMinuit->SetMaxIterations( 2000 );
  //rangemin5mm  = 50;  rangemax5mm  = 251;
  //rangemin10mm = 20;  rangemax10mm = 251;
  //startpar5mm[0] = 200;  startpar10mm[0] = 200;
  //startpar5mm[1] = 20;   startpar10mm[1] = 20;
  //startpar5mm[2] = 150;  startpar10mm[2] =  70;
  //startpar5mm[3] = -25;  startpar10mm[3] = -50;     
  //startpar5mm[4] = 100;  startpar10mm[4] = 100;
  //startpar5mm[5] = 0.02; startpar10mm[5] = 0.02;

  double startpar[nFitPar];
  startpar[0] = MaximumX();
  startpar[1] = 20; 
  startpar[2] = 20; 
  startpar[3] = 0.02;
  startpar[4] = 1;
  for ( int i = 0; i < nFitPar; i++ ) f1->SetParameter(i,startpar[i]);
  f1->SetParLimits(0,150,250) ;
  f1->SetParLimits(1,5,30) ;
  f1->SetParLimits(2,40,220) ;
  f1->SetParLimits(3,-100,-5) ;
  //f1->SetParLimits(4,1,10000) ;
  f1->SetParLimits(5,0,0.5) ;
  f1->SetRange(20,251) ;

  Fit("f1","RB") ;
  cout << "pars: " << f1->GetParameter(0) << " " << f1->GetParameter(1)
       << endl;
  Ssiz_t indx = gMinuit->fCstatu.Index("CONVERGED");
  //cout << "index = " << indx << endl;
  if ( indx >= 0 ) {
    T0 = -f1->GetParameter(0) - 2*f1->GetParameter(1);
    T0error = f1->GetParError(0);
  } else {
    cout << "** fit did not converge." << endl;
  }
  cout << "T0, T0error = " << T0 << " " << T0error << endl;
  delete f1;
  return OTrFObservable( T0, T0error );
}

#endif
