#ifndef __OTR_CHANNEL_HH__
#define __OTR_CHANNEL_HH__

#ifndef __OTR_COMPONENT_HH__
#include "OTrComponent.hh"
#endif

#ifndef __OTR_OBSERVABLE_HH__
#include "OTrObservable.hh"
#endif

#ifdef USE_ROOT
#ifndef ROOT_Rtypes
#include "Rtypes.h"
#endif
#ifndef ROOT_TString
#include <TString.h>
#endif
class TH1 ;
class TH1D ;
#define __OTR_HISTOGRAMMING__
#endif

class OTrEvent ;
class OTrWire ;
class OTrParameterizedRTRelation ;
class OTrRTRelation ;
class OTrHit ;
class OTrChannelHistogrammer ;

/** The channel class  **/
class OTrChannel : public OTrDaqComponent
{
public:
  /// typedefs
  typedef unsigned char  TdcCount_t ;
  typedef char           Mask_t  ;
  typedef OTrFObservable TZero_t ;

  /// default constructor
  OTrChannel();
  /// copy constructor
  OTrChannel(const OTrChannel&) ;
  /// use this one
  OTrChannel(ComponentId_t) ;
  /// destructor
  virtual ~OTrChannel() ;
  /// assignment operator
  OTrChannel& operator=(const OTrChannel&) ;

  // Setters and getters (for members only!)
  void SetWire(OTrWire* wire) { fWire=wire; } ;
  void SetTZero(const TZero_t& tzero) { fTZero=tzero; } ;
  void SetMask(const Mask_t mask) {fMask=mask ; } ;

  OTrWire* GetWire() const  { return fWire ; } ;
  Mask_t   GetMask() const  { return fMask ; } ;
  TZero_t  GetTZero() const { return fTZero ; } ;
#ifdef __OTR_HISTOGRAMMING__
  OTrChannelHistogrammer* GetHistogrammer() const { return fHistogrammer ; } ;
#endif

  /// process a new TDC count value. return hit if hit
  inline OTrHit* ProcessChannelBuffer( const TdcCount_t, OTrEvent*);
  /// return drifttime from tdc counts
  inline float GetDriftTime( const TdcCount_t count ) const ;
  /// inverse
  inline TdcCount_t GetTdcCount( const float driftTime) const ;
  /// for monte carlo we use a different t0
  inline TdcCount_t GetMCTdcCount( const float driftTime) const ;
  /// is the tdc value (and status) regarded as a hit or not 
  inline bool ValidHit( const TdcCount_t count ) const;
  // is the hit accepted; harder criteria 
  inline bool GoodHit( const TdcCount_t count ) const;
  /// is the mask accepted
  inline bool GoodMask() const;

  TdcCount_t GetMaximumTime() const ;
  TdcCount_t GetMinimumTime() const ;
  OTrParameterizedRTRelation* GetRTRelation() const ;
  OTrRTRelation* GetMCRTRelation() const ;
  void Reset() ;

#ifdef USE_ROOT
  ClassDef(OTrChannel,0) // The Outer Tracker channel class
  TString GetName() const ;
#endif

  // Statics
  static void SetRTRelation5mm(OTrParameterizedRTRelation* rtr) { fgRTRelation05mm=rtr; } ;
  static void SetRTRelation10mm(OTrParameterizedRTRelation* rtr) { fgRTRelation10mm=rtr; } ;
  static OTrParameterizedRTRelation* GetRTRelation5mm() { return fgRTRelation05mm ; } ;
  static OTrParameterizedRTRelation* GetRTRelation10mm() { return fgRTRelation10mm ; } ;
  static void SetMCRTRelation5mm(OTrRTRelation* rtr) { fgMCRTRelation05mm=rtr; } ;
  static void SetMCRTRelation10mm(OTrRTRelation* rtr) { fgMCRTRelation10mm=rtr; } ;
  static OTrRTRelation* GetMCRTRelation5mm() { return fgMCRTRelation05mm ; } ;
  static OTrRTRelation* GetMCRTRelation10mm() { return fgMCRTRelation10mm ; } ;

  // Forwarded to the histogrammer ...
  float GetOccupancy() const  ;
#ifdef __OTR_HISTOGRAMMING__
  float GetAverageTime() const ;
  float GetAverageTime2() const ;
  float GetSigma() const  ;
  int GetNHit() const  ;
  static void SetTimeHistogramming(bool b) ;
  static int GetNTimeBins()  ;
  static int GetMaximumTimeAxis() ;
  static int GetMinimumTimeAxis() ;
  static int GetHistogrammingMode()  ;
  static int Bin2Count( const int bin )  ;
  void PerformHistogramming(TdcCount_t tdc)  ;
  typedef unsigned short TimeHistogramBin_t ;
  const TimeHistogramBin_t* GetTimeHistogram();
  int GetTZeroBin() const  ;
  float GetAverageFromTimeH1() const  ;
#ifdef USE_ROOT
  virtual TH1D GetTimeH1() const ;
  TH1D* GetTimeH1_B( float tshift = 0 ) const  ;
#endif //USE_ROOT
#endif //__OTR_HISTOGRAMMING__

private:

  OTrWire* fWire ;  // pointer to corresponding wire
  Mask_t   fMask ;  // mask value, used to mark bad channels
  TZero_t  fTZero ; // time offset
#ifdef __OTR_HISTOGRAMMING__
  OTrChannelHistogrammer* fHistogrammer ; // pointer to histogrammer
#endif

  ///
  static TdcCount_t fgMinimumTime05mm ; // minumum drifttime for 05mm cells
  static TdcCount_t fgMaximumTime05mm ; // maximum drifttime for 05mm cells
  static TdcCount_t fgMinimumTime10mm ; // minumum drifttime for 10mm cells
  static TdcCount_t fgMaximumTime10mm ; // maximum drifttime for 10mm cells
  static Mask_t     fgMaximumMask ;     // maximum value for mask

  static OTrParameterizedRTRelation* fgRTRelation05mm  ; // rtrelation 05mm cells
  static OTrParameterizedRTRelation* fgRTRelation10mm ; // rtrelation 10mm cells
  static OTrRTRelation* fgMCRTRelation05mm ;  // rtrelation 05mm cells for MC
  static OTrRTRelation* fgMCRTRelation10mm ; // rtrelation 10mm cells for MC
};

ostream & operator << (ostream &, const OTrChannel& w );

/*************************** inlines **********************/

inline bool OTrChannel::GoodMask() const {
  //return GetMask() == 0;
  return (GetMask() < fgMaximumMask ); 
}

inline bool OTrChannel::GoodHit( const TdcCount_t count ) const
{
  return ( ValidHit( count ) && GoodMask() );
}

bool OTrChannel::ValidHit( const TdcCount_t count ) const
{
  return count <= GetMaximumTime() && 
    count >= GetMinimumTime() ;
}

inline float OTrChannel::GetDriftTime( const TdcCount_t count ) const {
  return count + GetTZero().GetValue() ;
};

inline OTrChannel::TdcCount_t 
OTrChannel::GetTdcCount( const float driftTime) const {
  return TdcCount_t(0.5 + driftTime - GetTZero().GetValue()) ;
};

inline OTrChannel::TdcCount_t 
OTrChannel::GetMCTdcCount( const float driftTime) const {
  return TdcCount_t(0.5 + driftTime) ;
};

#endif






