#ifndef __OTR_PARAMETERIZEDRTRELATION_HH__
#define __OTR_PARAMETERIZEDRTRELATION_HH__

#ifndef __SGI_STL_VECTOR
#include <vector>
#endif
#ifndef __OTR_FUNCTIONVALUE_HH__
#include "OTrFunctionValue.hh"
#endif
#ifndef __OTR_RTRELATION_HH__
#include "OTrRTRelation.hh"
#endif
#ifndef _Vector_H_
#include "CLHEP/Matrix/Vector.h"
#endif
#ifndef _SYMMatrix_H_
#include "CLHEP/Matrix/SymMatrix.h"
#endif

class TGraphErrors ;

/** This is the base class for parameterized rt relations */
class OTrParameterizedRTRelation : public OTrRTRelation
{
public:
  enum ParameterizedRTRelationType {kSpline,kChebychev} ;
  ///
  void UpdateTParameters(const HepVector& = 0, const HepSymMatrix& = 0) ;
  ///
  void UpdateSigma2Parameters(const HepVector& = 0, const HepSymMatrix& = 0) ;
  ///
  inline const HepVector& GetTParameters() const ;
  ///
  inline const HepSymMatrix& GetTCov() const ;
  ///
  inline const HepVector& GetSigma2Parameters() const ;
  ///
  inline const HepSymMatrix& GetSigma2Cov() const ;
  ///
  inline const int GetNoTParameters() const ;
  ///
  inline const int GetNoSigma2Parameters() const ;
  ///
  static void SetCorrectionLevel(int level=1) ;
  ///
  inline const HepVector& GetdTdPar(HepDouble) const ;
  ///
  inline const HepVector& GetdSigma2dPar(HepDouble) const ;
  ///
  HepVector CalcdTdPar(HepDouble) const ;
  ///
  HepVector CalcdSigma2dPar(HepDouble) const ;
  ///
  inline static ParameterizedRTRelationType GetType() ;
#ifdef USE_ROOT
  TGraphErrors *GetTGraph(int nPoints) ;
  TGraphErrors *GetSigmaGraph(int nPoints) ;
#endif
  const vector<HepVector>* GetdTdParTable() const { return &tdTdParTable ; } ;
private:
  ///
  OTrFunctionValue CalcFunctionValue(HepDouble,HepDouble,HepDouble,
				  const HepVector&) const ;
  ///
  HepVector CalcParDeriv(HepDouble,HepDouble,HepDouble,int) const ;
  ///
  HepVector fTParameters ;
  ///
  HepVector fSigma2Parameters ;
  ///
  HepSymMatrix fTCov ;
  ///
  HepSymMatrix fSigma2Cov ;
  ///
  vector<HepVector> tdTdParTable;
  ///
  vector<HepVector> tdSigma2dParTable;
protected:
  ///
  void SetTParameters(const HepVector&, const HepSymMatrix& = 0) ;
  ///
  void SetSigma2Parameters(const HepVector&, const HepSymMatrix& = 0) ;
  ///
  OTrParameterizedRTRelation() ;
  ///
  OTrParameterizedRTRelation(int noTpars, int noSigma2pars, double rmin, double rmax) ;
  ///
  virtual ~OTrParameterizedRTRelation() ;
  ///
  void FillBasisFuncTables() ;
  ///
  static int sCorrectionLevel;
  ///Declared by mother class
public:
  /// Read from inputstream
  virtual void Read(istream&) ;
  /// Writes the parameters to an outputstream
  virtual void Write(ostream&) const ;
  ///
  virtual OTrFunctionValue CalcFunctionT(const HepDouble R) const ;
  ///
  virtual OTrFunctionValue CalcFunctionSigma2(const HepDouble R) const ;
  ///
  virtual OTrRTResult CalcRTResult(const HepDouble R) const ;
  ///Implemented by child class
public:
  ///
  virtual vector<OTrFunctionValue>  CalcBasisFuncVector(HepDouble, int) const = 0;
  ///
  virtual void SetUniformSigma(HepDouble) = 0;
protected:
  ///
  virtual void CorrectT0() = 0;
  ///
  virtual void RepairTParameters() = 0;
  ///
  virtual void RepairSigma2Parameters() = 0;
  ///
  static const ParameterizedRTRelationType fgType ;
};


inline const HepVector&
OTrParameterizedRTRelation::GetTParameters() const {
  return fTParameters ; } ;

inline const HepVector&
OTrParameterizedRTRelation::GetSigma2Parameters() const {
  return fSigma2Parameters ; } ;

inline const HepSymMatrix&
OTrParameterizedRTRelation::GetTCov() const {
  return fTCov ; } ;

inline const HepSymMatrix&
OTrParameterizedRTRelation::GetSigma2Cov() const {
  return fSigma2Cov ; } ;

inline OTrFunctionValue
OTrParameterizedRTRelation::CalcFunctionT(const HepDouble R) const {
  return CalcFunctionValue(R,GetRMin(),GetRMax(),fTParameters) ; } ;

inline OTrFunctionValue
OTrParameterizedRTRelation::CalcFunctionSigma2(const HepDouble R) const {
  return CalcFunctionValue(R,GetRMin(),GetRMax(),fSigma2Parameters) ; 
} ;

inline const int
OTrParameterizedRTRelation::GetNoTParameters() const {
  return GetTParameters().num_row() ; } ;

inline const int
OTrParameterizedRTRelation::GetNoSigma2Parameters() const {
  return GetSigma2Parameters().num_row() ; } ;

inline const HepVector&
OTrParameterizedRTRelation::GetdSigma2dPar(const HepDouble R) const {
  const int noBins = tdSigma2dParTable.size();
  int rbin = int(noBins * (R-GetRMin())/(GetRMax()-GetRMin())) ;
  if(rbin<0) rbin=0;
  if(rbin>noBins-1) rbin=noBins-1 ;
  return tdSigma2dParTable[rbin] ;
};

inline const HepVector&
OTrParameterizedRTRelation::GetdTdPar(const HepDouble R) const {
  const int noBins = tdTdParTable.size();
  int rbin = int(noBins * (R-GetRMin())/(GetRMax()-GetRMin())) ;
  if(rbin<0) rbin=0;
  if(rbin>noBins-1) rbin=noBins-1 ;
  return tdTdParTable[rbin] ;
};

inline OTrParameterizedRTRelation::ParameterizedRTRelationType 
OTrParameterizedRTRelation::GetType() {
  return fgType;
};

#endif











